import io

import numpy as np
import torch
from PIL import Image
from skimage import draw


def compress_pil(im: np.ndarray, width: int = 512, quality=75):
    """im.shape = (H, W, C)"""
    if im.dtype != np.uint8:
        im = (np.clip(im, 0, 1) * 255).astype(np.uint8)

    pil_image = Image.fromarray(im)
    w, h = pil_image.size
    new_height = int(h * width / w)
    pil_image = pil_image.resize((width, new_height))
    buffer = io.BytesIO()
    pil_image.save(buffer, format="JPEG", quality=quality)
    buffer.seek(0)
    return Image.open(buffer)


def compute_relative_inset(box1, box2, shape=(1.0, 1.0)):
    """
    Compute the relative inset of box1 in box2 (wrt shape)
    """
    assert box1.shape == (2, 2) and box2.shape == (2, 2)

    if isinstance(box1, torch.Tensor):
        box1 = box1.detach().cpu().numpy()
    if isinstance(box2, torch.Tensor):
        box2 = box2.detach().cpu().numpy()

    b2_size = box2[1] - box2[0]
    scale = np.array(shape) / b2_size
    b1_rel = box1 - box2[0]
    b1_scaled = b1_rel * scale
    return b1_scaled


def box_annotate(x, bbox, lines=True):
    L, C, H, W = x.shape
    assert bbox.shape == (L, 2, 2)

    if isinstance(x, torch.Tensor):
        x = x.detach().cpu().numpy()
    if isinstance(bbox, torch.Tensor):
        bbox = bbox.detach().cpu().numpy()

    for i, (box1, box2) in enumerate(zip(bbox[:-1], bbox[1:])):
        box_rel = compute_relative_inset(box1, box2, shape=(H, W))
        rr, cc = draw.rectangle_perimeter(box_rel[0], box_rel[1], shape=(H, W))
        x[i + 1, :, rr, cc] = 1
        if lines:
            rr, cc, val = draw.line_aa(0, 0, int(box_rel[0][0]), int(box_rel[0][1]))
            x[i + 1, :, rr, cc] = np.repeat(val[:, None], C, axis=-1)
            rr, cc, val = draw.line_aa(0, W - 1, int(box_rel[0][0]), int(box_rel[1][1]))
            x[i + 1, :, rr, cc] = np.repeat(val[:, None], C, axis=-1)
    return x


def create_image_grid(x, output, bbox=None, lines=False, input_only=False, include_input=True, include_error=True, box_annotate_y=False, box_annotate_z=False):
    if include_error and not include_input:
        raise ValueError("include_error requires include_input to be True")
    # L, C, H, W
    if x.ndim == 5:
        B, L, C, H, W = x.shape
        D = None
        T = None
    elif x.ndim == 6:
        B, L, C, D, H, W = x.shape
        T = None
    elif x.ndim == 7:
        B, L, C, T, D, H, W = x.shape
    else:
        raise ValueError(f"Invalid input shape: {x.shape}")
    if include_input:
        x = x[0].detach().cpu().numpy()
    y = output["reco"][0].detach().cpu().numpy()
    z = output["input_masked"][0].detach().cpu().numpy()

    if T is not None and D is not None:
        if include_input:
            x = x[:, :, T // 2, D // 2]
        y = y[:, :, T // 2, D // 2]
        z = z[:, :, T // 2, D // 2]
        bbox = bbox[..., 2:]

    elif T is None and D is not None:
        if include_input:
            x = x[:, :, D // 2]
        y = y[:, :, D // 2]
        z = z[:, :, D // 2]
        bbox = bbox[..., 1:]
    if include_input and include_error:
        err = np.abs(x - y)


    if bbox is not None and bbox.shape[1] == L and bbox.shape[0] == 1:
        bbox = bbox[0] # Assume single sample, try remove batch dim

    if bbox is not None and include_input:
        box_annotate(x, bbox, lines=lines)

    if bbox is not None and box_annotate_y:
        box_annotate(y, bbox, lines=lines)
    
    if bbox is not None and box_annotate_z:
        box_annotate(z, bbox, lines=lines)

    if include_input:
        x = np.concatenate(
            np.pad(x, ((0, 0), (0, 0), (1, 1), (1, 1)), constant_values=1), axis=-2
        )
    y = np.concatenate(
        np.pad(y, ((0, 0), (0, 0), (1, 1), (1, 1)), constant_values=1), axis=-2
    )
    z = np.concatenate(
        np.pad(z, ((0, 0), (0, 0), (1, 1), (1, 1)), constant_values=1), axis=-2
    )
    if include_error:
        err = np.concatenate(
            np.pad(err, ((0, 0), (0, 0), (1, 1), (1, 1)), constant_values=1), axis=-2
        )
        out = np.concatenate([z, y, x, err], axis=-1)
    elif include_input and not input_only:
        out = np.concatenate([z, y, x], axis=-1)
    elif input_only:
        out = x
    else:
        out = np.concatenate([z, y], axis=-1)
    if out.shape[0] == 1:
        out = out[0]
    elif out.shape[0] == 2:
        # rgb
        out = np.stack([out[0], out[1], out[0]], axis=0)
    elif out.shape[0] > 3:
        out = out[:3]
    out = np.clip(out, 0, 1)
    return out



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Union

import lightning.pytorch as pl
import numpy as np
import torch
import wandb
from lightning.pytorch.utilities import grad_norm
from torch.optim.lr_scheduler import CosineAnnealingLR, LinearLR

from .utils import compress_pil, create_image_grid

if TYPE_CHECKING:
    from .mae import MuViTMAE


class WrappedModel(pl.LightningModule):
    def __init__(
        self,
        model: MuViTMAE,
        num_epochs: int,
        outdir: Union[Path, None],
        warmup_epochs: int = 5,
        learning_rate: float = 1e-4,
        nobox: bool = False,
        data_gen: torch.Generator = None,
    ):
        super().__init__()
        self.model = model
        self.max_epochs = num_epochs
        self.warmup_epochs = warmup_epochs
        self.learning_rate = learning_rate
        self.outdir = outdir
        self.nobox = nobox
        self.save_hyperparameters(ignore=["model"])
        self.data_gen = data_gen or torch.Generator()

    def on_train_batch_start(self, batch, batch_idx):
        # Access the learning rate from the optimizer
        optimizer = self.trainer.optimizers[0]  # Assuming one optimizer
        lr = optimizer.param_groups[0]["lr"]
        self.log("learning_rate", lr, on_step=True, on_epoch=False, prog_bar=False)

    def on_train_epoch_end(self):
        if self.trainer.is_global_zero and self.outdir is not None:
            self.model.save(self.outdir, overwrite=True)

    def on_train_epoch_start(self):
        epoch_seed = 42 + self.current_epoch
        self.data_gen.manual_seed(epoch_seed)

    def training_step(self, batch, batch_idx):
        if isinstance(batch, torch.Tensor) and batch.ndim == 5:
            bbox = None
        elif isinstance(batch, dict):
            bbox = batch.get("bbox", None)
            img = batch["img"]
        elif (
            len(batch) == 2
            and isinstance(batch[0], torch.Tensor)
            and batch[0].ndim == 5
        ):
            bbox = batch[1]
            img = batch[0]
        else:
            raise ValueError(
                f"Invalid batch format. Expected either a dictionary, a tuple (X,bbox) or a 5D tensor X, got {type(batch)}"
            )

        output = self.model(img, bbox=bbox if not self.nobox else None)
        loss = output["loss"]
        if torch.isnan(loss):
            print(f"Loss is NaN at step {self.global_step}")
            print(f"Batch shape: {img.shape}")
            print(f"Output: {output}")
            raise ValueError("Loss is NaN")

        self.log(
            "train_loss",
            loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            sync_dist=True,
            batch_size=len(img),
        )
        self.log("batch_size", len(img), on_step=True, prog_bar=False, sync_dist=False, batch_size=len(img))
        self.log(
            "batch_mean",
            img.mean().item(),
            on_step=True,
            prog_bar=False,
            sync_dist=False,
            batch_size=len(img),
        )
        _norm = grad_norm(self.model, 2).get("grad_2.0_norm_total", 0)
        self.log("grad_norm", _norm, on_step=True, prog_bar=False, sync_dist=False, batch_size=len(img))

        if (
            self.current_epoch % 2 == 0
            and self.trainer.is_global_zero
            and batch_idx == 0
            and not self.trainer.sanity_checking
        ):
            with torch.inference_mode():
                img = img[:1]
                if bbox is not None:
                    bbox = bbox[:1]
                output = self.model(img, bbox=bbox, return_all=True)
                self.log_images(img, output, bbox, batch_idx, name="train_inout")

        return loss

    def validation_step(self, batch, batch_idx):
        if isinstance(batch, torch.Tensor) and batch.ndim == 5:
            bbox = None
        elif isinstance(batch, dict):
            bbox = batch.get("bbox", None)
            img = batch["img"]
        elif (
            len(batch) == 2
            and isinstance(batch[0], torch.Tensor)
            and batch[0].ndim == 5
        ):
            bbox = batch[1]
            img = batch[0]
        else:
            raise ValueError(
                f"Invalid batch format. Expected either a dictionary, a tuple (X,bbox) or a 5D tensor X, got {type(batch)}"
            )
        output = self.model(img, bbox=bbox if not self.nobox else None, return_all=True)
        loss = output["loss"]
        self.log("val_loss", loss, on_epoch=True, prog_bar=True, sync_dist=True, batch_size=len(img))
        if (
            self.trainer.is_global_zero
            and batch_idx < 4
            and not self.trainer.sanity_checking
        ):
            self.log_images(img, output, bbox, batch_idx, name="val_inout")
            if self.current_epoch % 10 == 0:
                self.log_images(
                    img,
                    output,
                    bbox,
                    batch_idx,
                    name="val_bigger",
                    width=1024,
                    quality=95,
                )

        return loss

    def log_images(self, img, output, bbox, batch_idx, name, width=512, quality=95):
        out = create_image_grid(img, output, bbox[0] if bbox is not None else None)

        if isinstance(self.logger, pl.loggers.TensorBoardLogger):
            self.logger.experiment.add_images(
                f"{name}{batch_idx}", out[None], global_step=self.global_step
            )

        elif isinstance(self.logger, pl.loggers.WandbLogger):
            if out.ndim == 3:
                out = np.moveaxis(out, 0, -1)

            _img = wandb.Image(compress_pil(out, width=width, quality=quality))
            self.logger.experiment.log({f"{name}{batch_idx}": _img})

    def configure_optimizers(self):
        optimizer = torch.optim.AdamW(
            self.model.parameters(), betas=(0.9, 0.95), lr=self.learning_rate
        )
        scheduler = [
            {
                "scheduler": CosineAnnealingLR(optimizer, T_max=self.max_epochs),
            }
        ]
        scheduler.append(
            {
                "scheduler": LinearLR(
                    optimizer,
                    start_factor=1e-2,
                    end_factor=1.0,
                    total_iters=self.warmup_epochs,
                ),
            }
        )
        return [optimizer], scheduler



import os
os.environ["CUDA_VISIBLE_DEVICES"] = ""
import pytest
import torch
from muvit.encoders import MuViTEncoder2d, MuViTEncoder3d, MuViTEncoder4d
from tempfile import TemporaryDirectory
from pathlib import Path


NDIM_TO_ENCODER_CLS = {
    2: MuViTEncoder2d,
    3: MuViTEncoder3d,
    4: MuViTEncoder4d
}

def test_encoder_2d():
    encoder = MuViTEncoder2d(
        in_channels=1,
        levels=(1.0, 4.0),
        patch_size=(4, 4),
        num_layers=3,
        dim=64,
        heads=2,
        attention_mode="all",
        use_level_embed=True,
        rotary_mode="per_layer",
        input_space="real",
        dropout=0.0,
    )
    B = 2
    L = 2
    C = 1
    H = 32
    W = 32

    inp = torch.randn((B, L, C, H, W))
    with torch.inference_mode():
        _ = encoder(inp, return_intermediate_idxs=(0,2))

    C = 2
    inp = torch.randn((B, L, C, H, W))
    with pytest.raises(RuntimeError):
        with torch.inference_mode():
            _ = encoder(inp)


def test_encoder_3d():
    encoder = MuViTEncoder3d(
        in_channels=1,
        levels=(1.0, 4.0),
        patch_size=(4, 4, 4),
        num_layers=3,
        dim=64,
        heads=2,
        attention_mode="all",
        use_level_embed=True,
        rotary_mode="per_layer",
        input_space="real",
        dropout=0.0,
    )
    B = 2
    L = 2
    C = 1
    D = 16
    H = 32
    W = 32

    inp = torch.randn((B, L, C, D, H, W))
    with torch.inference_mode():
        _ = encoder(inp)

    C = 2
    inp = torch.randn((B, L, C, D, H, W))
    with pytest.raises(RuntimeError):
        with torch.inference_mode():
            _ = encoder(inp)


def test_encoder_4d():
    encoder = MuViTEncoder4d(
        in_channels=1,
        levels=(1.0, 4.0),
        patch_size=(4, 4, 4, 4),
        num_layers=3,
        dim=128,
        heads=2,
        attention_mode="all",
        use_level_embed=True,
        rotary_mode="per_layer",
        input_space="real",
        dropout=0.0,
    )
    B = 2
    L = 2
    C = 1
    T = 8
    D = 16
    H = 16
    W = 16

    inp = torch.randn((B, L, C, T, D, H, W))
    print(inp.shape)
    with torch.inference_mode():
        _ = encoder(inp)

    C = 2
    inp = torch.randn((B, L, C, T, D, H, W))
    with pytest.raises(RuntimeError):
        with torch.inference_mode():
            _ = encoder(inp)



@pytest.mark.parametrize("ndim", [2, 3, 4])
def test_encoder_io(ndim: int):
    ActualCls = NDIM_TO_ENCODER_CLS[ndim]
    OtherCls = NDIM_TO_ENCODER_CLS[(ndim+1) if ndim != 4 else 2]

    encoder = ActualCls(
        in_channels=1,
        levels=(1.0, 4.0),
        patch_size=(4,)*ndim,
        num_layers=3,
        dim=128,
        heads=2,
        attention_mode="all",
        use_level_embed=True,
        rotary_mode="per_layer",
        input_space="real",
        dropout=0.0,
    )
    with TemporaryDirectory() as tmpdir:
        encoder.save(Path(tmpdir)/"model", overwrite=True)

        # wrong ndim model
        with pytest.raises(ValueError):
            _ = OtherCls.from_folder(Path(tmpdir)/"model")

        loaded_encoder = ActualCls.from_folder(Path(tmpdir)/"model")
    assert encoder._config == loaded_encoder._config
    assert all((p1==p2).all() for p1, p2 in zip(encoder.parameters(), loaded_encoder.parameters()))


@pytest.mark.parametrize("rotary_mode", ["none", "fixed", "shared", "per_layer"])
def test_rotary_modes_2d(rotary_mode):
    encoder = MuViTEncoder2d(
        in_channels=1,
        levels=(1.0, 4.0),
        patch_size=(4, 4),
        num_layers=3,
        dim=64,
        heads=2,
        rotary_mode=rotary_mode,
    )
    B, L, C, H, W = 2, 2, 1, 32, 32
    inp = torch.randn((B, L, C, H, W))
    with torch.inference_mode():
        out, coords, level_idx = encoder(inp)
    assert out.shape == (B, L * (H // 4) * (W // 4), 64)

    rope_params = sum(p.numel() for n, p in encoder.named_parameters() if 'rotary_pos_emb' in n)
    if rotary_mode == "none":
        assert rope_params == 0
    elif rotary_mode == "fixed":
        assert rope_params == 0
    elif rotary_mode == "shared":
        assert rope_params == 16
    elif rotary_mode == "per_layer":
        assert rope_params == 48


@pytest.mark.parametrize("rotary_mode", ["none", "fixed", "shared", "per_layer"])
def test_rotary_sharing_2d(rotary_mode):
    encoder = MuViTEncoder2d(
        in_channels=1,
        levels=(1.0, 4.0),
        patch_size=(4, 4),
        num_layers=3,
        dim=64,
        heads=2,
        rotary_mode=rotary_mode,
    )

    if rotary_mode in ["fixed", "shared"]:
        assert id(encoder.layers[0].rotary_pos_embs) == id(encoder.layers[1].rotary_pos_embs)
        assert id(encoder.layers[1].rotary_pos_embs) == id(encoder.layers[2].rotary_pos_embs)
    elif rotary_mode == "per_layer":
        assert id(encoder.layers[0].rotary_pos_embs) != id(encoder.layers[1].rotary_pos_embs)
        assert id(encoder.layers[1].rotary_pos_embs) != id(encoder.layers[2].rotary_pos_embs)


def test_backward_compat_use_rotary_embed():
    """Test that old saved models with use_rotary_embed can still be loaded."""
    # Test use_rotary_embed=True -> rotary_mode=per_layer
    encoder_true = MuViTEncoder2d(
        in_channels=1,
        levels=(1.0, 4.0),
        patch_size=(4, 4),
        num_layers=3,
        dim=64,
        heads=2,
        rotary_mode="per_layer",
    )

    with TemporaryDirectory() as tmpdir:
        save_path = Path(tmpdir) / "model_true"
        encoder_true.save(save_path)

        # Manually modify config to use old use_rotary_embed parameter
        import yaml
        config_path = save_path / "config.yaml"
        with open(config_path, "r") as f:
            config = yaml.load(f, Loader=yaml.FullLoader)

        # Replace rotary_mode with use_rotary_embed
        config["use_rotary_embed"] = True
        config.pop("rotary_mode")

        with open(config_path, "w") as f:
            yaml.dump(config, f)

        # Test that loading converts use_rotary_embed to rotary_mode
        loaded_encoder = MuViTEncoder2d.from_folder(save_path)
        assert loaded_encoder._config["rotary_mode"] == "per_layer"
        assert "use_rotary_embed" not in loaded_encoder._config

    # Test use_rotary_embed=False -> rotary_mode=none
    encoder_false = MuViTEncoder2d(
        in_channels=1,
        levels=(1.0, 4.0),
        patch_size=(4, 4),
        num_layers=3,
        dim=64,
        heads=2,
        rotary_mode="none",
    )

    with TemporaryDirectory() as tmpdir:
        save_path = Path(tmpdir) / "model_false"
        encoder_false.save(save_path)

        import yaml
        config_path = save_path / "config.yaml"
        with open(config_path, "r") as f:
            config = yaml.load(f, Loader=yaml.FullLoader)

        # Replace rotary_mode with use_rotary_embed
        config["use_rotary_embed"] = False
        config.pop("rotary_mode")

        with open(config_path, "w") as f:
            yaml.dump(config, f)

        loaded_encoder = MuViTEncoder2d.from_folder(save_path)
        assert loaded_encoder._config["rotary_mode"] == "none"
        assert "use_rotary_embed" not in loaded_encoder._config


def test_extract_levels():
    """Test extracting subset of levels from encoder."""
    encoder = MuViTEncoder2d(
        in_channels=1,
        levels=(1.0, 2.0, 4.0),
        patch_size=(4, 4),
        num_layers=3,
        dim=64,
        heads=2,
        rotary_mode="per_layer",
        use_level_embed=True,
    )

    # Extract first two levels
    encoder_12 = encoder.extract_levels((1.0, 2.0))
    assert encoder_12.levels == (1.0, 2.0)
    assert len(encoder_12.proj) == 2
    assert encoder_12.level_embed.shape[0] == 2
    assert encoder_12._config["levels"] == (1.0, 2.0)

    # Extract single level
    encoder_1 = encoder.extract_levels((1.0,))
    assert encoder_1.levels == (1.0,)
    assert len(encoder_1.proj) == 1
    assert encoder_1.level_embed.shape[0] == 1

    # Extract in different order
    encoder_21 = encoder.extract_levels((2.0, 1.0))
    assert encoder_21.levels == (2.0, 1.0)

    # Test forward pass works
    x = torch.randn(2, 2, 1, 32, 32)
    out, coords, level_idx = encoder_12(x)
    assert out.shape[0] == 2
    assert out.shape[2] == 64


def test_extract_levels_validation():
    """Test that invalid levels raise errors."""
    encoder = MuViTEncoder2d(
        in_channels=1,
        levels=(1.0, 2.0, 4.0),
        patch_size=(4, 4),
        num_layers=3,
        dim=64,
        heads=2,
    )

    # Invalid level should raise error
    with pytest.raises(ValueError, match="Level 3.0 not found"):
        encoder.extract_levels((1.0, 3.0))


def test_extract_levels_copy():
    """Test copy=False behavior."""
    encoder = MuViTEncoder2d(
        in_channels=1,
        levels=(1.0, 2.0, 4.0),
        patch_size=(4, 4),
        num_layers=3,
        dim=64,
        heads=2,
    )

    # With copy=False and same levels, should return self
    encoder_same = encoder.extract_levels((1.0, 2.0, 4.0), copy=False)
    assert encoder_same is encoder

    # With copy=True (default), should return new object
    encoder_copy = encoder.extract_levels((1.0, 2.0, 4.0))
    assert encoder_copy is not encoder
    assert encoder_copy.levels == encoder.levels


def test_extract_levels_save_reload(tmp_path):
    """Test that extracted encoder can be saved and reloaded."""
    encoder = MuViTEncoder2d(
        in_channels=1,
        levels=(1.0, 2.0, 4.0),
        patch_size=(4, 4),
        num_layers=3,
        dim=64,
        heads=2,
        rotary_mode="per_layer",
        use_level_embed=True,
    )

    # Extract subset of levels
    encoder_12 = encoder.extract_levels((1.0, 2.0))

    # Save extracted encoder
    save_path = tmp_path / "encoder_12"
    encoder_12.save(save_path)

    # Reload and verify
    loaded = MuViTEncoder2d.from_folder(save_path)
    assert loaded.levels == (1.0, 2.0)
    assert len(loaded.proj) == 2
    assert loaded.level_embed.shape[0] == 2
    assert loaded._config["levels"] == (1.0, 2.0)

    # Verify forward pass works
    x = torch.randn(2, 2, 1, 32, 32)
    out_original, _, _ = encoder_12(x)
    out_loaded, _, _ = loaded(x)
    assert torch.allclose(out_original, out_loaded, atol=1e-6)


if __name__ == "__main__":
    test_encoder_4d()
    print("4d fwd ok")
    test_encoder_io(4)
    print("4d io ok")


import os
os.environ["CUDA_VISIBLE_DEVICES"] = ""
import pytest
import torch
from muvit.data import MuViTDataset
from pathlib import Path

from typing import Tuple

class DummyDataset(MuViTDataset):
    def __init__(
        self,
        num_samples: int,
        n_levels: int,
        spatial_size: Tuple[int, int, int],
        n_channels: int,
    ):
        super().__init__()
        self.num_samples = num_samples
        self.spatial_size = spatial_size
        self._levels = list(range(1, n_levels+1))
        self._n_channels = n_channels
        self._ndim = len(spatial_size)

    @property
    def levels(self):
        return self._levels

    @property
    def n_channels(self):
        return self._n_channels

    @property
    def ndim(self):
        return self._ndim

    def __len__(self):
        return self.num_samples

    def __getitem__(self, idx: int):
        # Generate a random image and a random mask
        if idx > len(self):
            raise IndexError("Dataset index out of range")
        img = (
            torch.randn((self.n_levels, self.n_channels, *self.spatial_size))
            .float()
        )
        bbox = torch.randn((self.n_levels, 2, self.ndim)).float()
        return {
            "img": img,
            "bbox": bbox,
        }
    
class BuggyDummyDataset(DummyDataset):
    def __init__(self, bug_type: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._bug_type = bug_type
    
    def __getitem__(self, idx: int):
        res = super().__getitem__(idx)
        if self._bug_type == "img_shape":
            res["img"] = res["img"].unsqueeze(0)
        elif self._bug_type == "img_lvs":
            res["img"] = torch.stack(2*[res["img"]], dim=0)
        elif self._bug_type == "img_ch":
            res["img"] = torch.stack(2*[res["img"]], dim=0)
        elif self._bug_type == "bbox_shape":
            res["bbox"] = res["bbox"].unsqueeze(0)
        elif self._bug_type == "bbox_lvs":
            res["bbox"] = torch.stack(2*[res["bbox"]], dim=0)
        elif self._bug_type == "bbox_ndim":
            res["bbox"] = torch.stack(2*[res["bbox"]], dim=2)
        return res

@pytest.mark.parametrize("ndim", [2, 3, 4])
def test_data_ok(ndim: int):
    ndim_to_spatial_size = {
        2: (32, 32),
        3: (16, 32, 32),
        4: (4, 16, 32, 32)
    }
    ds = DummyDataset(
        num_samples=10,
        n_levels=2,
        spatial_size=ndim_to_spatial_size[ndim],
        n_channels=1,
    )
    _item = ds[3]
    assert "img" in _item.keys()
    assert "bbox" in _item.keys()

@pytest.mark.parametrize("ndim", [2, 3, 4])
@pytest.mark.parametrize("bug_type", ["img_shape", "img_lvs", "img_ch", "bbox_shape", "bbox_lvs", "bbox_ndim"])
def test_data_buggy(ndim: int, bug_type: str):
    ndim_to_spatial_size = {
        2: (32, 32),
        3: (16, 32, 32),
        4: (4, 16, 32, 32)
    }
    with pytest.raises(ValueError):
        _ = BuggyDummyDataset(
            bug_type=bug_type,
            num_samples=10,
            n_levels=2,
            spatial_size=ndim_to_spatial_size[ndim],
            n_channels=1,
        )

if __name__ == "__main__":
    ds = DummyDataset(10, 2, (1,32,32),1)


from abc import ABC, ABCMeta, abstractmethod
from pathlib import Path
from typing import Literal, Optional

from torch.utils.data import Dataset

import matplotlib.pyplot as plt
from skimage.color import label2rgb


class SanityCheckMeta(ABCMeta):
    def __call__(cls, *args, **kwargs):
        obj = super().__call__(*args, **kwargs)
        if hasattr(obj, "_sanity_check"):
            obj._sanity_check()
        return obj


class MuViTDataset(Dataset, ABC, metaclass=SanityCheckMeta):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @abstractmethod
    def __getitem__(self, idx: int):
        pass

    @abstractmethod
    def __len__(self):
        pass

    @property
    @abstractmethod
    def ndim(self):
        pass

    @property
    def n_levels(self):
        return len(self.levels)

    @property
    @abstractmethod
    def levels(self):
        pass

    @property
    @abstractmethod
    def n_channels(self):
        pass

    def _sanity_check(self):
        if self.ndim not in (2, 3, 4):
            raise ValueError("ndim must be 2, 3 or 4")
        if len(self) == 0:
            raise ValueError("Dataset is empty.")
        sample = self[0]
        if "img" not in sample or "bbox" not in sample:
            raise ValueError("Sample must contain 'img' and 'bbox' keys.")

        img = sample["img"]
        bbox = sample["bbox"]

        if bbox is not None:
            expected_bbox_shape = (self.n_levels, 2, self.ndim)
            if tuple(bbox.shape) != expected_bbox_shape:
                raise ValueError(
                    f"Expected bbox shape {expected_bbox_shape}, got {tuple(bbox.shape)}"
                )

        expected_img_ndim = {2: 4, 3: 5, 4: 6}[self.ndim]
        if img.ndim != expected_img_ndim:
            raise ValueError(
                f"Expected img.ndim = {expected_img_ndim} for ndim={self.ndim} (L,C"
                f"{',D' if self.ndim == 3 else ''},H,W), got {img.ndim}"
            )

        if img.shape[0] != self.n_levels:
            raise ValueError(
                f"Expected img.shape[0] (L) = {self.n_levels}, got {img.shape[0]}"
            )
        if img.shape[1] != self.n_channels:
            raise ValueError(
                f"Expected img.shape[1] (C) = {self.n_channels}, got {img.shape[1]}"
            )

    def visualize_sample(self, idx: int = 0, save_file: Optional[Path] = None, view: Literal["yx", "zx", "zy"] = "yx", continuous_label_cmap=None):
        if self.ndim not in (2,3):
            raise NotImplementedError(
                "Sample visualization is implemented for 2d and 3d only."
            )
        import numpy as np

        from .utils import box_annotate

        _item = self[idx]

        img, bbox = _item["img"].cpu().numpy(), _item["bbox"].cpu().numpy()
        if self.ndim == 3 and view not in ("yx", "zy", "zx"):
            raise ValueError("view must be one of 'yx', 'zy', 'zx'")

        if self.ndim == 3 and view == "yx":
            img = img[..., img.shape[-3] // 2, :, :]
            bbox = bbox[:, :, 1:]
        elif self.ndim == 3 and view == "zx":
            img = img[..., img.shape[-2] // 2, :]
            bbox = bbox[:, :, [0, 2]]
        elif self.ndim == 3 and view == "zy":
            img = img[..., img.shape[-1] // 2]
            bbox = bbox[:, :, :2]


        box_annotate(img, bbox)
        img = np.concatenate(
            np.pad(img, ((0, 0), (0, 0), (1, 1), (1, 1)), constant_values=1), axis=-2
        )

        if img.shape[0] == 1:
            img = np.repeat(img, 3, axis=0)
        elif img.shape[0] == 2:
            img = np.stack([img[0], img[1], img[0]], axis=0)
        elif img.shape[0] > 3:
            img = img[:3]

        if "label" in _item and _item["label"] is not None:
            label = _item["label"].cpu().numpy()
            
            if self.ndim == 3 and view == "yx":
                label = label[..., label.shape[-3] // 2, :, :]
            elif self.ndim == 3 and view == "zx":
                label = label[..., label.shape[-2] // 2, :]
            elif self.ndim == 3 and view == "zy":
                label = label[..., label.shape[-1] // 2]

            label_cp = np.zeros((label.shape[0], 3, label.shape[1], label.shape[2]))



            if continuous_label_cmap is None:
                for s in range(label.shape[0]):
                    label_cp[s] = label2rgb(label[s], channel_axis=0) # CYX
            else:
                cmap_fun = plt.cm.get_cmap(continuous_label_cmap)
                for s in range(label.shape[0]):
                    label_cp[s] = cmap_fun(label[s])[...,:3].transpose(2,0,1) # C,Y,X

            box_annotate(label_cp, bbox)
            label_cp = np.concatenate(np.pad(label_cp, ((0, 0), (0, 0), (1, 1), (1, 1)), constant_values=1),axis=-2)
            if img.ndim != label_cp.ndim:
                img = np.stack([img]*label_cp.shape[0], axis=0)
            img = np.concatenate([img, label_cp], axis=-1)

        if save_file:
            save_file = Path(save_file)
            save_file.parent.mkdir(exist_ok=True, parents=True)
            fig, ax = plt.subplots(figsize=(10, 10))
            if img.ndim > 2:
                ax.imshow(img.transpose(1,2,0))
            else:
                ax.imshow(img, cmap="gray")
            ax.axis("off")
            ax.spines["top"].set_visible(False)
            ax.spines["right"].set_visible(False)
            ax.spines["left"].set_visible(False)
            ax.spines["bottom"].set_visible(False)
            fig.savefig(save_file, bbox_inches="tight")
            plt.close(fig)



import os

os.environ["CUDA_VISIBLE_DEVICES"] = ""
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Tuple

import pytest
import torch
from muvit.data import MuViTDataset
from muvit.mae import MuViTMAE2d, MuViTMAE3d, MuViTMAE4d

NDIM_TO_MAE_CLS = {
    2: MuViTMAE2d,
    3: MuViTMAE3d,
    4: MuViTMAE4d
}

class DummyDataset(MuViTDataset):
    def __init__(
        self,
        num_samples: int,
        n_levels: int,
        spatial_size: Tuple[int, int, int, int],
        n_channels: int,
    ):
        super().__init__()
        self.num_samples = num_samples
        self.spatial_size = spatial_size
        self._levels = list(range(1, n_levels+1))
        self._n_channels = n_channels
        self._ndim = len(spatial_size)

    @property
    def levels(self):
        return self._levels

    @property
    def n_channels(self):
        return self._n_channels

    @property
    def ndim(self):
        return self._ndim

    def __len__(self):
        return self.num_samples

    def __getitem__(self, idx: int):
        # Generate a random image and a random mask
        if idx > len(self):
            raise IndexError("Dataset index out of range")
        img = (
            torch.randn((self.n_levels, self.n_channels, *self.spatial_size))
            .float()
        )
        bbox = torch.randn((self.n_levels, 2, self.ndim)).float()
        return {
            "img": img,
            "bbox": bbox,
        }


@pytest.mark.parametrize("in_channels", [1, 3])
@pytest.mark.parametrize("ndim", [2, 3, 4])
def test_mae_fwd(in_channels: int, ndim: int):
    if ndim not in (2, 3, 4):
        raise ValueError("ndim must be either 2, 3, or 4.")
    ActualCls = NDIM_TO_MAE_CLS[ndim]

    model = ActualCls(
        in_channels=in_channels,
        levels=(1, 4),
        num_layers=3,
        heads=2,
        masking_ratio=0.1,
        dim=320,
        dim_decoder=256,
        patch_size=4,
    )
    B = 1
    L = 2
    C = in_channels
    H = 32
    W = 32
    if ndim == 2:
        inp = torch.randn((B, L, C, H, W))
    elif ndim == 3:
        D = 16
        inp = torch.randn((B, L, C, D, H, W))
    elif ndim == 4:
        D = 16
        T = 4
        inp = torch.randn((B, L, C, T, D, H, W))
    with torch.inference_mode():
        _ = model(inp)


@pytest.mark.parametrize("in_channels", [1, 3])
@pytest.mark.parametrize("ndim", [2, 3, 4])
@pytest.mark.filterwarnings("ignore:Starting from*") # lightning logging warning
@pytest.mark.filterwarnings("ignore:The '*") # lightning dataloader warning
def test_mae_fit(in_channels: int, ndim: int):
    if ndim not in (2, 3, 4):
        raise ValueError("ndim must be either 2 or 3.")
    ActualCls = ActualCls = NDIM_TO_MAE_CLS[ndim]

    model = ActualCls(
        in_channels=in_channels,
        levels=(1, 4),
        num_layers=3,
        heads=2,
        masking_ratio=0.1,
        dim=320,
        dim_decoder=256,
        input_space="real",
        patch_size=4,
    )
    dim_to_spatial_size = {
        2: (32,32),
        3: (16,32,32),
        4: (4,16,32,32)
    }
    train_ds = DummyDataset(
        num_samples=16,
        n_levels=2,
        spatial_size=dim_to_spatial_size[ndim],
        n_channels=in_channels,
    )
    val_ds = DummyDataset(
        num_samples=4,
        n_levels=2,
        spatial_size=dim_to_spatial_size[ndim],
        n_channels=in_channels,
    )
    train_dl = torch.utils.data.DataLoader(train_ds, batch_size=4, shuffle=True)
    val_dl = torch.utils.data.DataLoader(val_ds, batch_size=1, shuffle=False)
    model.fit(train_dl, val_dl, output=None, dry=True, fast_dev_run=True)


@pytest.mark.parametrize("ndim", [2, 3, 4])
def test_mae_io(ndim: int):
    ActualCls = NDIM_TO_MAE_CLS[ndim]
    OtherCls = NDIM_TO_MAE_CLS[(ndim+1) if ndim != 4 else 2]
    model = ActualCls(
        in_channels=1,
        levels=(1, 4),
        num_layers=3,
        heads=2,
        masking_ratio=0.1,
        dim=128,
        dim_decoder=128,
        input_space="real",
        patch_size=4,
    )
    with TemporaryDirectory() as tmpdir:
        model.save(Path(tmpdir) / "model", overwrite=True)

        # Load wrong ndim model
        with pytest.raises(ValueError):
            _ = OtherCls.from_folder(Path(tmpdir) / "model")

        loaded_model = ActualCls.from_folder(Path(tmpdir) / "model")
    assert model._config == loaded_model._config
    assert all(
        (p1 == p2).all()
        for p1, p2 in zip(model.parameters(), loaded_model.parameters())
    )

if __name__ == "__main__":
    test_mae_fwd("real", 1, 4)



import logging
import sys
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import Generic, Literal, Optional, Tuple, Type, TypeVar, Union

import lightning.pytorch as pl
import numpy as np
import torch
import torch.nn.functional as F
from argparse import Namespace
from einops import rearrange
from torch import Tensor, nn

from .bblocks import SaveableModel
from .decoders import MuViTDecoder, MuViTDecoder2d, MuViTDecoder3d, MuViTDecoder4d
from .encoders import MuViTEncoder, MuViTEncoder2d, MuViTEncoder3d, MuViTEncoder4d
from .trainer import WrappedModel

T = TypeVar("T", bound=Tuple[int, ...])


log = logging.getLogger(__name__)
log.setLevel(logging.INFO)

console_handler = logging.StreamHandler(sys.stdout)
console_handler.setLevel(logging.INFO)
formatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")
console_handler.setFormatter(formatter)
log.addHandler(console_handler)


class MuViTMAE(SaveableModel, ABC, Generic[T]):
    def __init__(
        self,
        in_channels: int = 1,
        levels: tuple[float] = (1, 4),
        patch_size: Union[int, tuple[int, ...]] = 16,
        num_layers: int = 12,
        dim: int = 512,
        num_layers_decoder: int = 2,
        dim_decoder: Optional[int] = 256,
        heads: int = 8,
        decoder_mode: Literal["single", "multi", "multi_iso"] = "multi",
        loss: Literal["mse", "norm_mse"] = "mse",
        masking_ratio: float = 0.75,
        use_level_embed: bool = True,
        rotary_mode: Literal["none", "fixed", "shared", "per_layer"] = "per_layer",
        rotary_base: int = 10000,
        attention_mode: Literal["all", "causal", "same", "random"] = "all",
        masking_mode: Literal["dirichlet", "random"] | tuple[float] = "dirichlet",
        input_space: Literal["real"] = "real",
        dropout: float = 0.0,
    ):
        """Initialize a Masked Autoencoder with multi-level Vision Transformer.

        Args:
            in_channels: Number of input channels
            levels: Tuple of scale factors for each level
            patch_size: Size of image patches
            num_layers: Number of encoder layers
            dim: Hidden dimension
            num_layers_decoder: Number of decoder layers
            dim_decoder: Hidden dimension for decoder
            heads: Number of attention heads
            decoder_mode: Type of decoder ('single', 'multi', 'multi_iso')
            loss: Type of loss function ('mse', 'norm_mse')
            masking_ratio: Ratio of patches to mask
            use_level_embed: Whether to use level embeddings
            rotary_mode: Type of rotary embeddings (none/fixed/shared/per_layer)
            rotary_base: Base for rotary embeddings
            attention_mode: Type of attention masking
            masking_mode: Type of masking ('dirichlet', 'random' or tuple of probabilities)
            dropout: Dropout probability for transformer layers
        """
        super().__init__()

        if dim_decoder is None:
            dim_decoder = dim

        # Select appropriate encoder/decoder based on spatial dimensionality
        EncoderCls = self.encoder_class
        DecoderCls = self.decoder_class

        self._config = dict(
            in_channels=in_channels,
            levels=levels,
            patch_size=patch_size,
            num_layers=num_layers,
            dim=dim,
            heads=heads,
            use_level_embed=use_level_embed,
            rotary_mode=rotary_mode,
            rotary_base=rotary_base,
            decoder_mode=decoder_mode,
            loss=loss,
            masking_ratio=masking_ratio,
            masking_mode=masking_mode,
            attention_mode=attention_mode,
            num_layers_decoder=num_layers_decoder,
            dim_decoder=dim_decoder,
            dropout=dropout,
            input_space=input_space,
            ndim=self.ndim,
        )

        self.encoder = EncoderCls(
            in_channels=in_channels,
            levels=levels,
            patch_size=patch_size,
            num_layers=num_layers,
            dim=dim,
            heads=heads,
            use_level_embed=use_level_embed,
            rotary_mode=rotary_mode,
            rotary_base=rotary_base,
            attention_mode=attention_mode,
            dropout=dropout,
            input_space=input_space,
        )

        # Calculate patch dimension for final layer
        patch_dim = in_channels * np.prod(self.encoder.patch_size)
        self.final = nn.Linear(dim_decoder, patch_dim)

        self.patch_size = self.encoder.patch_size
        self.masking_ratio = masking_ratio
        self.masking_mode = masking_mode
        self.decoder_mode = decoder_mode

        if decoder_mode == "single":
            self.mask_token = nn.Parameter(torch.randn(1, 1, dim))
            self.decoder = DecoderCls(
                dim,
                dim_decoder,
                num_layers=num_layers_decoder,
                heads=heads,
                rotary_mode=rotary_mode,
                rotary_base=rotary_base,
                dropout=dropout,
            )
        elif decoder_mode in ("multi", "multi_iso"):
            self.mask_token = nn.Parameter(
                torch.randn(1, len(self.encoder.levels), dim)
            )
            self.decoder = nn.ModuleList(
                [
                    DecoderCls(
                        dim,
                        dim_decoder,
                        num_layers=num_layers_decoder,
                        heads=heads,
                        rotary_mode=rotary_mode,
                        rotary_base=rotary_base,
                        dropout=dropout,
                    )
                    for _ in range(len(self.encoder.levels))
                ]
            )
        else:
            raise ValueError(f"Invalid decoder mode: {decoder_mode}")
        self.loss_fn = loss

    @classmethod
    @property
    @abstractmethod
    def encoder_class(self) -> Type[MuViTEncoder]:
        """Get the appropriate encoder class for this dimensionality."""
        pass

    @classmethod
    @property
    @abstractmethod
    def decoder_class(self) -> Type[MuViTDecoder]:
        """Get the appropriate decoder class for this dimensionality."""
        pass

    @classmethod
    @property
    @abstractmethod
    def ndim(self) -> int:
        """Get the number of dimensions for this model."""
        pass

    @abstractmethod
    def patch_token_to_image(
        self, x: Tensor, *shape: int, space_transform: bool = True
    ) -> Tensor:
        """Convert patch tokens back to image."""
        pass

    def token_to_patch(self, x: Tensor) -> Tensor:
        if self.ndim == 2:
            return rearrange(
                x,
                "b n (p1 p2) -> b n p1 p2",
                p1=self.patch_size[0],
                p2=self.patch_size[1],
            )
        elif self.ndim == 3:
            return rearrange(
                x,
                "b n (p1 p2 p3) -> b n p1 p2 p3",
                p1=self.patch_size[0],
                p2=self.patch_size[1],
                p3=self.patch_size[2],
            )
        elif self.ndim == 4:
            return rearrange(
                x,
                "b n (p1 p2 p3 p4) -> b n p1 p2 p3 p4",
                p1=self.patch_size[0],
                p2=self.patch_size[1],
                p3=self.patch_size[2],
                p4=self.patch_size[3],
            )
        else:
            raise ValueError(f"Invalid number of dimensions: {self.ndim}")

    def forward(
        self,
        x: Tensor,
        bbox: Optional[tuple[Tensor, Tensor]] = None,
        return_all: bool = False,
        eps: float = 1e-2,
        consistent_levels: bool=False,
        rng_generator: Optional[torch.Generator]=None,
        masking_mode_is_ratio: bool=False,
    ):
        """Process input through the MAE model.

        Args:
            x: Input tensor
            bbox: Optional bounding box coordinates
            return_all: Whether to return intermediate results
            eps: Small constant for numerical stability
            consistent_levels: Whether to use consistent masking levels across the batch. If True, the same token locations will be masked out for different runs (assuming same RNG), otherwise different runs may have different masked tokens. Defaults to False.
            rng_generator: Optional random number generator for reproducibility of the `consistent_levels` option.
            masking_mode_is_ratio: Whether the masking_mode is given as a ratio (True) denoting the explicit fraction of tokens to mask, or as a tuple of probabilities (False).

        Returns:
            Dictionary containing:
            - input: Original patches
            - output: Reconstructed patches
            - loss: Reconstruction loss
            - encoded: Encoded features (if return_all)
            - reco: Reconstructed image (if return_all)
            - decoded: Decoded features (if return_all)
            - input_masked: Masked input (if return_all)
            - loss_per_level: Loss per level (if return_all)
        """
        y, coords, patches, batch_range, idx_retain, idx_mask = (
            self.encoder.forward_masked(x, bbox, self.masking_ratio, self.masking_mode, consistent_levels, rng_generator, masking_mode_is_ratio=masking_mode_is_ratio)
        )

        N = patches.shape[1]
        N_per_level = N // len(self.encoder.levels)

        z = torch.zeros(
            patches.shape[0], N, y.shape[-1], device=patches.device, dtype=patches.dtype
        )

        if self.decoder_mode == "single":
            z[batch_range, idx_mask] = self.mask_token
            z[batch_range, idx_retain] = y
            # single self-attention for all levels
            z = self.decoder(z, coords)
        elif self.decoder_mode in ("multi", "multi_iso"):
            mask_tokens = torch.repeat_interleave(
                self.mask_token, N_per_level, dim=1
            ).repeat(patches.shape[0], 1, 1)
            z[batch_range, idx_mask] = mask_tokens[batch_range, idx_mask]
            z[batch_range, idx_retain] = y
            zs = torch.split(z, N_per_level, dim=1)
            cs = torch.split(coords, N_per_level, dim=1)
            if self.decoder_mode == "multi":
                # allow cross attending to all levels
                z = torch.cat(
                    [
                        self.decoder[i](_z, _c, context=z, context_coords=coords)
                        for i, (_z, _c) in enumerate(zip(zs, cs))
                    ],
                    dim=1,
                )
            elif self.decoder_mode == "multi_iso":
                # do it in isolation for each level
                z = torch.cat(
                    [self.decoder[i](_z, _c) for i, (_z, _c) in enumerate(zip(zs, cs))],
                    dim=1,
                )
            else:
                raise ValueError(f"Invalid decoder mode: {self.decoder_mode}")
        else:
            raise ValueError(f"Invalid decoder mode: {self.decoder_mode}")

        z = self.final(z)

        # normalize target
        if self.loss_fn == "norm_mse":
            p_mean = patches.mean(dim=-1, keepdim=True)
            p_std = patches.std(dim=-1, keepdim=True)
        elif self.loss_fn in ("mse", "mse_fft"):
            p_mean = torch.tensor(0, device=x.device, dtype=patches.dtype)
            p_std = torch.tensor(1 - eps, device=x.device, dtype=patches.dtype)
        else:
            raise ValueError(f"Invalid loss: {self.loss_fn}")

        patches_normed = (patches - p_mean) / (p_std + eps)

        loss = F.mse_loss(
            z[batch_range, idx_mask], patches_normed[batch_range, idx_mask]
        )
        if self.loss_fn == "mse_fft":
            z2 = self.token_to_patch(z[batch_range, idx_mask]).to(torch.float32)
            patches2 = self.token_to_patch(patches_normed[batch_range, idx_mask]).to(
                torch.float32
            )
            z2f = torch.fft.rfftn(z2, dim=tuple(-(i + 1) for i in range(self.ndim)))
            patches2f = torch.fft.rfftn(
                patches2, dim=tuple(-(i + 1) for i in range(self.ndim))
            )
            loss = loss + 0.01 * F.l1_loss(z2f, patches2f)

        out = dict(input=patches, output=z, loss=loss)
        if return_all:
            reco = patches_normed.clone().to(z.dtype)
            reco[batch_range, idx_mask] = z[batch_range, idx_mask]
            reco = reco * (p_std + eps) + p_mean
            reco = self.encoder.patch_space_transform(reco, inverse=True)
            reco = self.patch_token_to_image(reco, *x.shape[2:]).to(patches.dtype)

            input_masked = patches.clone()
            input_masked = self.encoder.patch_space_transform(
                input_masked, inverse=True
            )
            input_masked[batch_range, idx_mask] = 0.5
            input_masked = self.patch_token_to_image(input_masked, *x.shape[2:])
            mask = torch.zeros(patches.shape, dtype=bool, device=x.device)
            mask[batch_range, idx_mask] = True
            mask = self.patch_token_to_image(mask, *x.shape[2:])
            
            # loss per lv
            B = idx_mask.shape[0]
            loss_per_level_list = []
            for i in range(len(self.encoder.levels)):
                sel = (idx_mask // N_per_level) == i  # (B, M_mask)

                selected_z = []
                selected_p = []
                for b in range(B):
                    sel_b = sel[b]
                    if sel_b.any():
                        idxs = idx_mask[b, sel_b]
                        selected_z.append(z[b, idxs])
                        selected_p.append(patches_normed[b, idxs])

                if len(selected_z) == 0:
                    loss_per_level_list.append(torch.tensor(0.0, device=x.device))
                else:
                    zz = torch.cat(selected_z, dim=0)
                    pp = torch.cat(selected_p, dim=0)
                    loss_per_level_list.append(F.mse_loss(zz, pp))

            loss_per_level = torch.stack(loss_per_level_list).to(x.device)

            out["encoded"] = y
            out["reco"] = reco
            out["decoded"] = z
            out["mask"] = mask
            out["input_masked"] = input_masked
            out["loss_per_level"] = loss_per_level
        return out

    def fit(
        self,
        train_dataloader,
        val_dataloader,
        output: Union[Path, str, None],
        num_epochs: int = 500,
        lr: Optional[float] = None,
        nobox: bool = False,
        warmup_epochs: int = 2,
        logger: Optional[str] = "wandb",
        run_name: Optional[str] = None,
        wandb_project: Optional[str] = "muvit",
        ckpt_path: Optional[Union[str, Path]] = None,
        accelerator: Optional[str] = "auto",
        num_nodes: Optional[int] = 1,
        precision: Optional[str] = "bf16-mixed",
        strategy: Optional[str] = "auto",
        data_gen: Optional[torch.Generator] = None,
        dry: Optional[bool] = False,
        gradient_clip_val: Optional[float] = 0.5,
        fast_dev_run: bool = False,
        args_namespace: Optional[Namespace] = None
    ):
        if num_nodes < 1:
            raise ValueError("The number of nodes should be at least 1.")

        if logger in ("tensorboard", "tb"):
            logger = pl.loggers.TensorBoardLogger(save_dir=output, name=run_name)
        elif logger == "wandb":
            logger = pl.loggers.WandbLogger(
                project=wandb_project, name=run_name, save_dir=output
            )
            if args_namespace is not None:
                logger.log_hyperparams(vars(args_namespace))
        elif logger is None:
            if not dry:
                log.warning(
                    "No logger provided. Training will run, but progress will not be logged."
                )
        else:
            raise ValueError(f"Invalid logger: {logger}")

        if run_name is None:
            run_name = datetime.now().strftime("%Y%m%d_%H%M%S")

        if lr is None:
            n_gpus = max(torch.cuda.device_count(), 1)
            lr = 1e-4 * np.sqrt(train_dataloader.batch_size * n_gpus * num_nodes / 128)

        wrapped_model = WrappedModel(self, num_epochs, output, warmup_epochs, lr, nobox, data_gen)

        trainer = pl.Trainer(
            max_epochs=num_epochs,
            accelerator=accelerator,
            precision=precision,
            strategy=strategy,
            num_nodes=num_nodes,
            gradient_clip_val=gradient_clip_val,
            default_root_dir=output if not dry else None,
            logger=logger if not dry else None,
            fast_dev_run=fast_dev_run,
        )
        trainer.fit(
            wrapped_model, train_dataloader, val_dataloader, ckpt_path=ckpt_path
        )
        return

    def extract_levels(self, levels: tuple[float, ...], copy: bool = True) -> "MuViTMAE":
        """Return a new MuViTMAE instance restricted to a subset of levels. Follows the same signature as MuViTEncoder.extract_levels().

        Args:
            levels: Tuple of level scales to keep (must be subset of self.encoder.levels)
            copy: If False and levels match self.encoder.levels, return self without copying

        Returns:
            A new MuViTMAE instance with encoder/decoder restricted to ``levels``.
        """
        if not copy and tuple(levels) == tuple(self.encoder.levels):
            return self

        for level in levels:
            if level not in tuple(self.encoder.levels):
                raise ValueError(f"Requested level {level} not present in encoder levels {self.encoder.levels}")

        level_indices = tuple(self.encoder.levels.index(level) for level in levels)

        new_config = self._config.copy()
        new_config["levels"] = levels
        new_config.pop("ndim", None)

        new_mae = type(self)(**new_config)

        try:
            new_mae.final.load_state_dict(self.final.state_dict())
        except Exception:
            print("Warning: Could not copy final projection layer weights to new MAE.")
            pass

        new_encoder = self.encoder.extract_levels(levels, copy=True)
        new_mae.encoder = new_encoder
       
        if self.decoder_mode == "single":
            new_mae.decoder.load_state_dict(self.decoder.state_dict())
            new_mae.mask_token = nn.Parameter(self.mask_token.data.clone())
        else:
            for new_idx, old_idx in enumerate(level_indices):
                new_mae.decoder[new_idx].load_state_dict(
                    self.decoder[old_idx].state_dict()
                )

            if self.mask_token is not None:
                new_mae.mask_token = nn.Parameter(self.mask_token.data[:, level_indices, :].clone())
        return new_mae
   

class MuViTMAE2d(MuViTMAE[Tuple[int, int]]):
    @classmethod
    @property
    def ndim(self) -> int:
        return 2

    @classmethod
    @property
    def encoder_class(self) -> Type[MuViTEncoder]:
        return MuViTEncoder2d

    @classmethod
    @property
    def decoder_class(self) -> Type[MuViTDecoder]:
        return MuViTDecoder2d

    def patch_token_to_image(self, x: Tensor, C: int, H: int, W: int) -> Tensor:
        """Convert patch tokens back to image.

        Input: (B, N, P*P*C) -> Output: (B, L, C, H, W)
        """
        x = torch.split(x, x.shape[1] // len(self.encoder.levels), dim=1)
        x = torch.stack(
            [
                rearrange(
                    _x,
                    "b (h w) (p1 p2 c) -> b c (h p1) (w p2)",
                    c=C,
                    h=H // self.patch_size[0],
                    w=W // self.patch_size[1],
                    p1=self.patch_size[0],
                    p2=self.patch_size[1],
                )
                for _x in x
            ],
            dim=1,
        )
        return x


class MuViTMAE3d(MuViTMAE[Tuple[int, int, int]]):
    @classmethod
    @property
    def ndim(self) -> int:
        return 3

    @classmethod
    @property
    def encoder_class(self) -> Type[MuViTEncoder]:
        return MuViTEncoder3d

    @classmethod
    @property
    def decoder_class(self) -> Type[MuViTDecoder]:
        return MuViTDecoder3d

    def patch_token_to_image(self, x: Tensor, C: int, D: int, H: int, W: int) -> Tensor:
        """Convert patch tokens back to image.

        Input: (B, N, D*H*W*C) -> Output: (B, L, C, D, H, W)
        """
        x = torch.split(x, x.shape[1] // len(self.encoder.levels), dim=1)
        x = torch.stack(
            [
                rearrange(
                    _x,
                    "b (d h w) (p1 p2 p3 c) -> b c (d p1) (h p2) (w p3)",
                    c=C,
                    d=D // self.patch_size[0],
                    h=H // self.patch_size[1],
                    w=W // self.patch_size[2],
                    p1=self.patch_size[0],
                    p2=self.patch_size[1],
                    p3=self.patch_size[2],
                )
                for _x in x
            ],
            dim=1,
        )
        return x


class MuViTMAE4d(MuViTMAE[Tuple[int, int, int]]):
    @classmethod
    @property
    def ndim(self) -> int:
        return 4

    @classmethod
    @property
    def encoder_class(self) -> Type[MuViTEncoder]:
        return MuViTEncoder4d

    @classmethod
    @property
    def decoder_class(self) -> Type[MuViTDecoder]:
        return MuViTDecoder4d

    def patch_token_to_image(self, x: Tensor, C: int, T: int, D: int, H: int, W: int) -> Tensor:
        """Convert patch tokens back to image.

        Input: (B, N, T*D*H*W*C) -> Output: (B, L, C, T, D, H, W)
        """
        x = torch.split(x, x.shape[1] // len(self.encoder.levels), dim=1)
        x = torch.stack(
            [
                rearrange(
                    _x,
                    "b (t d h w) (p1 p2 p3 p4 c) -> b c (t p1) (d p2) (h p3) (w p4)",
                    c=C,
                    t=T // self.patch_size[0],
                    d=D // self.patch_size[1],
                    h=H // self.patch_size[2],
                    w=W // self.patch_size[3],
                    p1=self.patch_size[0],
                    p2=self.patch_size[1],
                    p3=self.patch_size[2],
                    p4=self.patch_size[3],
                )
                for _x in x
            ],
            dim=1,
        )
        return x


import pytest
import torch
from muvit.decoders import MuViTDecoder2d, MuViTDecoder3d, MuViTDecoder4d


def test_decoder_2d():
    decoder = MuViTDecoder2d(
        in_channels=64,
        dim=64,
        num_layers=3,
        heads=2,
        dropout=0.0,
    )
    B = 2
    N = 128
    D = 64

    inp = torch.randn((B, N, D))
    with torch.inference_mode():
        _ = decoder(inp)


def test_decoder_3d():
    decoder = MuViTDecoder3d(
        in_channels=64,
        dim=64,
        num_layers=3,
        heads=2,
        dropout=0.0,
    )
    B = 2
    N = 128
    D = 64

    inp = torch.randn((B, N, D))
    with torch.inference_mode():
        _ = decoder(inp)

def test_decoder_4d():
    decoder = MuViTDecoder4d(
        in_channels=64,
        dim=128,
        num_layers=3,
        heads=2,
        dropout=0.0,
    )
    B = 2
    N = 128
    D = 64

    inp = torch.randn((B, N, D))
    with torch.inference_mode():
        _ = decoder(inp)



from abc import ABC, abstractmethod
from typing import Generic, Literal, Optional, Tuple, TypeVar

import torch.nn as nn
from torch import Tensor
from x_transformers.x_transformers import RotaryEmbedding

from .bblocks import SaveableModel, TransformerLayer, RotaryEmbeddingTrainable, _split_rope_dims

T = TypeVar("T", bound=Tuple[int, ...])

class MuViTDecoder(SaveableModel, ABC, Generic[T]):
    def __init__(
        self,
        in_channels: int,
        dim: int = 320,
        num_layers: int = 2,
        heads: int = 4,
        rotary_mode: Literal["none", "fixed", "shared", "per_layer"] = "per_layer",
        rotary_base: int = 10000,
        dropout: float = 0.0,
        attention_mode: Literal["all", "causal", "same", "random"] = "all",
    ):
        """Initialize a Vision Transformer decoder.

        Args:
            in_channels: Number of input channels
            dim: Hidden dimension
            num_layers: Number of transformer layers
            heads: Number of attention heads
            rotary_mode: Type of rotary embeddings (none/fixed/shared/per_layer)
            rotary_base: Base for rotary embeddings
            dropout: Dropout probability for transformer layers
        """
        super().__init__()

        self.proj = nn.Sequential(
            nn.Linear(in_channels, dim), nn.GELU(), nn.LayerNorm(dim)
        )

        shared_rope = None
        per_layer_rotary_dim = None

        dims = _split_rope_dims(dim, heads, self.ndim)

        if rotary_mode == "none":
            pass
        elif rotary_mode == "fixed":
            shared_rope = nn.ModuleList([
                RotaryEmbedding(dim=d, base=rotary_base) for d in dims
            ])
        elif rotary_mode == "shared":
            shared_rope = nn.ModuleList([
                RotaryEmbeddingTrainable(dim=d, base=rotary_base) for d in dims
            ])
        elif rotary_mode == "per_layer":
            per_layer_rotary_dim = self.ndim
        else:
            raise ValueError(
                f"Invalid rotary_mode: '{rotary_mode}'. "
                f"Must be one of: 'none', 'fixed', 'shared', 'per_layer'"
            )

        if shared_rope is not None:
            self.rotary_pos_embs = shared_rope

        self.layers = nn.ModuleList(
            [
                TransformerLayer(
                    dim=dim,
                    heads=heads,
                    rotary_dim=per_layer_rotary_dim,
                    rotary_pos_embs=shared_rope,
                    rotary_base=rotary_base,
                    dropout=dropout,
                )
                for _ in range(num_layers)
            ]
        )
        self.dim = dim
        self.attention_mode = attention_mode

    def forward(
        self,
        x: Tensor,
        coords: Optional[Tensor] = None,
        context: Optional[Tensor] = None,
        context_coords: Optional[Tensor] = None,
    ):
        """Decode features with optional cross-attention.

        Args:
            x: Input features of shape (B, N, D)
            coords: Coordinates for rotary embeddings
            context: Optional context for cross-attention
            context_coords: Coordinates for context rotary embeddings

        Returns:
            Decoded features of same shape as input
        """
        B, _, D = x.shape
        x = self.proj(x)
        if context is not None:
            context = self.proj(context)

        for i, layer in enumerate(self.layers):
            # use cross attention for first layer
            if i == 0:
                x = layer(
                    x, coords=coords, context=context, context_coords=context_coords, attention_mode=self.attention_mode
                )
            else:
                x = layer(x, coords=coords, attention_mode=self.attention_mode)
        return x
    
    @classmethod
    @property
    @abstractmethod
    def ndim(self) -> int:
        """Number of spatial dimensions"""
        pass


class MuViTDecoder2d(MuViTDecoder[Tuple[int, int]]):
    @classmethod
    @property
    def ndim(self) -> int: # FIXME: deprecated in py313
        return 2

class MuViTDecoder3d(MuViTDecoder[Tuple[int, int, int]]):
    @classmethod
    @property
    def ndim(self) -> int: # FIXME: deprecated in py313
        return 3

class MuViTDecoder4d(MuViTDecoder[Tuple[int, int, int]]):
    @classmethod
    @property
    def ndim(self) -> int: # FIXME: deprecated in py313
        return 4



from abc import ABC, abstractmethod
from typing import Generic, Literal, Optional, Tuple, TypeVar, Union

import numpy as np
import torch
from einops import rearrange
from torch import Tensor, nn
from x_transformers.x_transformers import RotaryEmbedding

from .bblocks import SaveableModel, TransformerLayer, RotaryEmbeddingTrainable, _split_rope_dims

T = TypeVar("T", bound=Tuple[int, ...])


class MuViTEncoder(SaveableModel, ABC, Generic[T]):
    def __init__(
        self,
        in_channels: int,
        levels: tuple[float],
        patch_size: Union[int, tuple[int, ...]],
        num_layers: int = 6,
        dim: int = 320,
        heads: int = 4,
        attention_mode: Literal["all", "causal", "same", "random"] = "all",
        use_level_embed: bool = True,
        rotary_mode: Literal["none", "fixed", "shared", "per_layer"] = "per_layer",
        rotary_base: int = 10000,
        input_space: Literal["real"] = "real",
        dropout: float = 0.0,
    ):
        """Initialize a multi-level Vision Transformer encoder.

        Args:
            in_channels: Number of input channels
            levels: Tuple of scale factors for each level
            patch_size: Size of image patches
            num_layers: Number of transformer layers
            dim: Hidden dimension
            heads: Number of attention heads
            attention_mode: Type of attention masking
            use_level_embed: Whether to use level embeddings
            rotary_mode: Type of rotary embeddings (none/fixed/shared/per_layer)
            rotary_base: Base for rotary embeddings
            input_space: Input space (real)
            dropout: Dropout probability
        """
        super().__init__()

        self.levels = levels
        self.patch_size = (
            patch_size if isinstance(patch_size, tuple) else (patch_size,) * self.ndim
        )
        self.dim = dim
        self.attention_mode = attention_mode
        self.in_channels = in_channels

        self._config = dict(
            in_channels=in_channels,
            levels=levels,
            patch_size=patch_size,
            num_layers=num_layers,
            dim=dim,
            heads=heads,
            attention_mode=attention_mode,
            use_level_embed=use_level_embed,
            rotary_mode=rotary_mode,
            rotary_base=rotary_base,
            dropout=dropout,
            input_space=input_space,
            ndim=self.ndim,
        )

        self.input_space = input_space
        patch_dim = in_channels * np.prod(self.patch_size)
        self.proj = nn.ModuleList(
            [
                nn.Sequential(nn.Linear(patch_dim, dim), nn.LayerNorm(dim))
                for _ in range(len(levels))
            ]
        )

        if use_level_embed:
            self.level_embed = nn.Parameter(torch.randn(len(levels), 1, 1, dim))
        else:
            self.level_embed = None

        shared_rope = None
        per_layer_rotary_dim = None

        dims = _split_rope_dims(dim, heads, self.ndim)

        if rotary_mode == "none":
            pass
        elif rotary_mode == "fixed":
            shared_rope = nn.ModuleList([
                RotaryEmbedding(dim=d, base=rotary_base) for d in dims
            ])
        elif rotary_mode == "shared":
            shared_rope = nn.ModuleList([
                RotaryEmbeddingTrainable(dim=d, base=rotary_base) for d in dims
            ])
        elif rotary_mode == "per_layer":
            per_layer_rotary_dim = self.ndim
        else:
            raise ValueError(
                f"Invalid rotary_mode: '{rotary_mode}'. "
                f"Must be one of: 'none', 'fixed', 'shared', 'per_layer'"
            )

        if shared_rope is not None:
            self.rotary_pos_embs = shared_rope

        self.layers = nn.ModuleList(
            [
                TransformerLayer(
                    dim=dim,
                    heads=heads,
                    rotary_dim=per_layer_rotary_dim,
                    rotary_pos_embs=shared_rope,
                    rotary_base=rotary_base,
                    dropout=dropout,
                )
                for _ in range(num_layers)
            ]
        )

    @classmethod
    @property
    @abstractmethod
    def ndim(self) -> int:
        """Number of spatial dimensions."""
        pass

    def patchify(self, x: Tensor) -> Tensor:
        """Convert input tensor to patches.

        Input: (B,C,(T),(D),H,W) -> Output: (B,N,(P*)(P*)P*P*C)
        """
        if self.ndim == 2:
            # B, C, H, W
            x = rearrange(
                x,
                "b c (h p1) (w p2) -> b (h w) (p1 p2 c)",
                p1=self.patch_size[0],
                p2=self.patch_size[1],
            )
        elif self.ndim == 3:
            # B, C, D, H, W
            x = rearrange(
                x,
                "b c (d p1) (h p2) (w p3) -> b (d h w) (p1 p2 p3 c)",
                p1=self.patch_size[0],
                p2=self.patch_size[1],
                p3=self.patch_size[2],
            )
        elif self.ndim == 4:
            # B, C, T, D, H, W
            x = rearrange(
                x,
                "b c (t p1) (d p2) (h p3) (w p4) -> b (t d h w) (p1 p2 p3 p4 c)",
                p1=self.patch_size[0],
                p2=self.patch_size[1],
                p3=self.patch_size[2],
                p4=self.patch_size[3],
            )
        else:
            raise ValueError(f"Invalid number of dimensions: {self.ndim}")
        return x

    @abstractmethod
    def get_patch_coords(
        self, B: int, L: int, *shape: int, bbox: Optional[tuple[Tensor, Tensor]] = None
    ) -> Tensor:
        """Get coordinates for each patch."""
        pass

    @abstractmethod
    def patch_embed(
        self, x: Tensor, bbox: Optional[tuple[Tensor, Tensor]] = None
    ) -> Tuple[Tensor, Tensor, Tensor]:
        """Process multi-level input through patch embedding."""
        pass

    def patch_space_transform(self, patches: Tensor, inverse: bool = False) -> Tensor:
        """transform if input space is not real."""
        if self.input_space == "real":
            return patches
        else:
            raise ValueError(f"Invalid input space: {self.input_space}")

    def forward_masked(
        self,
        x: Tensor,
        bbox: Optional[Tensor] = None,
        masking_ratio: float = 0.75,
        masking_mode: Literal["dirichlet", "random"] | tuple[float] = "dirichlet",
        consistent_levels: bool = False,
        generator: Optional[torch.Generator] = None,
        masking_mode_is_ratio: bool = False,
    ):
        x, patches, coords = self.patch_embed(x, bbox)

        B, N, D = x.shape
        if not N % len(self.levels) == 0:
            raise ValueError(
                f"Number of patches must be divisible by number of levels: {N} % {len(self.levels)} != 0"
            )

        N_per_level = N // len(self.levels)
        
        level_idx = (
            torch.arange(patches.shape[1], device=x.device)
            .unsqueeze(0)
            .repeat(patches.shape[0], 1)
        )
        level_idx = level_idx // N_per_level

        if masking_mode_is_ratio and isinstance(masking_mode, (tuple, list, np.ndarray)):
            if not len(masking_mode) == len(self.levels):
                raise ValueError(
                    f"Number of levels must match number of masking ratios: {len(masking_mode)} != {len(self.levels)}"
                )
            
            # explicit per-level masking ratios
            all_indices = []
            for b in range(B):
                retained_indices = []
                for level_idx_val, level_mask_ratio in enumerate(masking_mode):
                    level_start = level_idx_val * N_per_level
                    level_end = (level_idx_val + 1) * N_per_level
                    level_patches = torch.arange(level_start, level_end, device=x.device)
                    
                    n_retain_level = int(N_per_level * (1 - level_mask_ratio))
                    
                    if n_retain_level > 0:
                        perm = torch.randperm(N_per_level, device=x.device, generator=generator)
                        retained_patches = level_patches[perm[:n_retain_level]]
                        retained_indices.append(retained_patches)
                
                if retained_indices:
                    retained_indices_cat = torch.cat(retained_indices, dim=0)
                else:
                    retained_indices_cat = torch.tensor([], dtype=torch.long, device=x.device)
                
                all_patches = torch.arange(N, device=x.device)
                masked_indices = all_patches[~torch.isin(all_patches, retained_indices_cat)]
                
                batch_idx = torch.cat([retained_indices_cat, masked_indices], dim=0)
                all_indices.append(batch_idx)
            
            idx = torch.stack(all_indices, dim=0)
            N_retained = len(retained_indices_cat) if retained_indices else 0
        else:
            N_retained = int(N * (1 - masking_ratio))
            
            if masking_mode == "dirichlet":
                # alpha_parameter = 1.0
                alpha_parameter = 0.5
                prob_weights = torch.distributions.Dirichlet(
                    torch.ones(len(self.levels), device=x.device) * alpha_parameter
                ).sample()
            elif masking_mode == "random":
                prob_weights = torch.ones(len(self.levels), device=x.device)
            elif isinstance(masking_mode, (tuple, list, np.ndarray)):
                if not len(masking_mode) == len(self.levels):
                    raise ValueError(
                        f"Number of levels must match number of masking probabilities: {len(masking_mode)} != {len(self.levels)}"
                    )
                prob_weights = torch.tensor(masking_mode, device=x.device)
            else:
                raise ValueError(f"Invalid masking mode: {masking_mode}")

            if consistent_levels:
                idx = self.consistent_level_mask_sampling(
                    B=B,
                    N=N,
                    N_per_level=N_per_level,
                    masking_ratio=masking_ratio,
                    device=x.device,
                    generator=generator,
                )
            else:
                prob_per_block = torch.repeat_interleave(prob_weights, N_per_level)
                idx = torch.stack(
                    [torch.multinomial(prob_per_block, N, replacement=False) for _ in range(B)],
                    dim=0,
                )

        idx_retain, idx_mask = idx[:, :N_retained], idx[:, N_retained:]

        batch_range = torch.arange(B, device=x.device)[:, None]
        x = x[batch_range, idx_retain]
        level_idx = level_idx[batch_range, idx_retain]

        for layer in self.layers:
            x = layer(
                x,
                level_idx=level_idx,
                coords=coords[batch_range, idx_retain],
                attention_mode=self.attention_mode,
            )

        return x, coords, patches, batch_range, idx_retain, idx_mask

    def consistent_level_mask_sampling(
        self,
        B: int,
        N: int,
        N_per_level: int,
        masking_ratio: float,
        device: torch.device,
        generator: Optional[torch.Generator] = None,
    ) -> Tensor:
        """
        Deterministically sample retained/masked indices per level.

        The selection for level l depends only on (generator, batch_idx, level)
        so that runs with different sets of levels will produce identical masks
        for common levels when using the same generator at the same status.

        Make sure no shuffling of data happens between different runs and that the 
        order of calls is preserved.
        """

        L = N // N_per_level
        
        base_seed = int(torch.randint(0, 2**31 - 1, (1,), generator=generator).item())

        retained_per_level = max(1, int(round(N_per_level * (1.0 - masking_ratio))))

        all_idx = []
        for b in range(B):
            retained_globals = []
            for i_lvl in range(L):
                gen = torch.Generator(device=device)
                level_value = self.levels[i_lvl]
                level_hash = hash(level_value) & 0x7FFFFFFF  # always > 0
                seed = base_seed + b * 424242 + level_hash
                gen.manual_seed(int(seed))
                perm = torch.randperm(N_per_level, generator=gen, device=device)
                kept = perm[:retained_per_level] + i_lvl * N_per_level
                retained_globals.append(kept)

            retained_globals = torch.cat(retained_globals, dim=0)

            mask = torch.zeros(N, dtype=torch.bool, device=device)
            mask[retained_globals] = True
            rest = torch.arange(N, device=device)[~mask]
            idx_row = torch.cat([retained_globals, rest], dim=0)
            if idx_row.numel() != N:
                idx_row = idx_row[:N]
                if idx_row.numel() < N:
                    missing = torch.arange(N, device=device)[~torch.isin(torch.arange(N, device=device), idx_row)]
                    idx_row = torch.cat([idx_row, missing], dim=0)
            all_idx.append(idx_row)

        return torch.stack(all_idx, dim=0)

    def forward(self, x: Tensor, bbox: Optional[Tensor] = None, return_intermediate_idxs: Optional[Tuple[int]] = None):
        """Process multi-level input through the encoder."""
        x, patches, coords = self.patch_embed(x, bbox)
        B, N, D = x.shape

        N_per_level = N // len(self.levels)

        level_idx = (
            torch.arange(patches.shape[1], device=x.device)
            .unsqueeze(0)
            .repeat(patches.shape[0], 1)
        )
        level_idx = level_idx // N_per_level
        if return_intermediate_idxs is not None:
            interm_outputs = []
        for i, layer in enumerate(self.layers):
            x = layer(
                x,
                level_idx=level_idx,
                coords=coords,
                attention_mode=self.attention_mode,
            )
            if return_intermediate_idxs is not None and i in return_intermediate_idxs:
                interm_outputs.append(x)

        if return_intermediate_idxs is not None:
            return x, coords, level_idx, tuple(interm_outputs)
        else:
            return x, coords, level_idx

    def extract_levels(
        self, levels: tuple[float, ...], copy: bool = True
    ) -> "MuViTEncoder":
        """Extract subset of levels from this encoder.

        Args:
            levels: Level scales to keep (must be subset of self.levels)
            copy: If False and levels match self.levels, return self without copying

        Returns:
            Encoder with specified levels (new object unless copy=False and levels unchanged)
        """
        # Check if no-op
        if not copy and tuple(levels) == self.levels:
            return self

        # Validate requested levels exist
        for level in levels:
            if level not in self.levels:
                raise ValueError(
                    f"Level {level} not found in encoder. "
                    f"Available levels: {self.levels}"
                )

        # Map requested levels to indices
        level_indices = tuple(self.levels.index(level) for level in levels)

        # Create new config with filtered levels
        new_config = self._config.copy()
        new_config["levels"] = levels
        # Remove ndim as it's a class property, not a constructor parameter
        new_config.pop("ndim", None)

        # Create new encoder
        new_encoder = type(self)(**new_config)

        # Copy weights for selected levels
        for new_idx, old_idx in enumerate(level_indices):
            new_encoder.proj[new_idx].load_state_dict(
                self.proj[old_idx].state_dict()
            )
            if self.level_embed is not None:
                new_encoder.level_embed.data[new_idx] = self.level_embed.data[old_idx]

        # Copy shared components
        new_encoder.layers.load_state_dict(self.layers.state_dict())
        if hasattr(self, "rotary_pos_embs"):
            new_encoder.rotary_pos_embs.load_state_dict(
                self.rotary_pos_embs.state_dict()
            )

        return new_encoder


class MuViTEncoder2d(MuViTEncoder[Tuple[int, int]]):
    @classmethod
    @property
    def ndim(self) -> int: # FIXME: breaks in py313 as classmethod+properties are deprecated
        return 2

    def get_patch_coords(
        self,
        B: int,
        L: int,
        H: int,
        W: int,
        bbox: Optional[tuple[Tensor, Tensor]] = None,
    ) -> Tensor:
        """Get coordinates for each patch.

        Args:
            B: Batch size
            L: Number of levels
            H: Height of input
            W: Width of input
            bbox: Optional bounding box coordinates defining the spatial extent of each crop.
                 Shape (B, L, 2, 2) where dims are (batch, level, top/bottom, y/x).
        Returns: (B,L*H*W,2) coordinates
        """
        h, w = H // self.patch_size[0], W // self.patch_size[1]

        if bbox is None:
            bbox = (
                [[-l * (H / 2 - 0.5), -l * (W / 2 - 0.5)] for l in self.levels],
                [[l * (H / 2 - 0.5), l * (W / 2 - 0.5)] for l in self.levels],
            )
            bbox = torch.tensor(bbox).permute(1, 0, 2)
            bbox = bbox.view(1, L, 2, 2).repeat(B, 1, 1, 1)

        if not bbox.ndim == 4 and bbox.shape[:3] == (B, L, 2):
            raise ValueError(
                f"Bounding box must be of shape (B, L, 2, 2): {bbox.shape}"
            )

        coords = torch.meshgrid(
            torch.linspace(0, 1, h), torch.linspace(0, 1, w), indexing="ij"
        )
        coords = torch.stack(coords, dim=-1).to(bbox.device)  # (h, w, 2)

        coords = coords.view(1, 1, h, w, 2)  # (1, 1, h, w, 2)
        coords = coords.repeat(B, L, 1, 1, 1)  # (B, L, h, w, 2)

        # Scale and offset coords according to bbox
        top_left = bbox[:, :, 0, :].view(B, L, 1, 1, 2)  # (B, L, 1, 1, 2)
        bottom_right = bbox[:, :, 1, :].view(B, L, 1, 1, 2)  # (B, L, 1, 1, 2)
        coords = top_left + coords * (bottom_right - top_left)  # (B, L, h, w, 2)

        coords = rearrange(coords, "b l h w c -> b (l h w) c")
        return coords.to(bbox.device)

    def patch_embed(
        self, x: Tensor, bbox: Optional[tuple[Tensor, Tensor]] = None
    ) -> Tuple[Tensor, Tensor, Tensor]:
        """Process multi-level input through patch embedding.

        Input: x (B, L, C, H, W), bbox (B, L, 2, 2)
        Output: x (B, N, D), patches (B, N, C*P*P), coords (B, N, 2)
        """
        B, L, C, H, W = x.shape
        assert L == len(self.levels), "Number of levels must match"

        if bbox is not None:
            expected_shape = (B, L, 2, 2)
            if not (bbox.ndim == 4 and bbox.shape[:3] == (B, L, 2)):
                raise ValueError(
                    f"Bounding box must be of shape {expected_shape}: {bbox.shape}"
                )

        coords = self.get_patch_coords(B, L, H, W, bbox=bbox).to(x.device)

        # patchify assumes batch dimension is first, so we loop over levels
        x = x.transpose(0, 1)
        patches = tuple(self.patch_space_transform(self.patchify(_x)) for _x in x)
        x = tuple(pro(pat) for pro, pat in zip(self.proj, patches))

        if self.level_embed is not None:
            x = tuple(x + le for x, le in zip(x, self.level_embed))

        # back to B, L order
        x = torch.cat(x, dim=1)
        patches = torch.cat(patches, dim=1)

        return x, patches, coords

    def compute_features(self, x: Tensor, bbox: Optional[Tensor] = None) -> Tensor:
        """
        Compute features from a given input stack/bbox and return them in a spatially structured way.

        Args:
            x: Input tensor of shape (B, L, C, H, W)
            bbox: Optional bounding box coordinates of shape (B, L, 2, 2)

        Returns:
            feats: Features of shape (B, L, D, H', W') where H' = H // patch_size[0], W' = W // patch_size[1], D is the feature dimension
        """
        feats, _, _ = self(x, bbox=bbox)
        H, W = x.shape[-2], x.shape[-1]
        L = len(self.levels)
        h = H // self.patch_size[0]
        w = W // self.patch_size[1]
        return rearrange(feats, "b (l h w) d -> b l d h w", l=L, h=h, w=w)


class MuViTEncoder3d(MuViTEncoder[Tuple[int, int, int]]):
    @classmethod
    @property
    def ndim(self) -> int:
        return 3

    def get_patch_coords(
        self,
        B: int,
        L: int,
        D: int,
        H: int,
        W: int,
        bbox: Optional[tuple[Tensor, Tensor]] = None,
    ) -> Tensor:
        """Get coordinates for each patch.

        Returns: (B,L*D*H*W,3) coordinates
        """
        d, h, w = (
            D // self.patch_size[0],
            H // self.patch_size[1],
            W // self.patch_size[2],
        )

        if bbox is None:
            bbox = (
                [
                    [-l * (D / 2 - 0.5), -l * (H / 2 - 0.5), -l * (W / 2 - 0.5)]
                    for l in self.levels
                ],
                [
                    [l * (D / 2 - 0.5), l * (H / 2 - 0.5), l * (W / 2 - 0.5)]
                    for l in self.levels
                ],
            )
            bbox = torch.tensor(bbox).permute(1, 0, 2)
            bbox = bbox.view(1, L, 2, 3).repeat(B, 1, 1, 1)

        if not bbox.ndim == 4 and bbox.shape[:3] == (B, L, 2):
            raise ValueError(
                f"Bounding box must be of shape (B, L, 2, 3): {bbox.shape}"
            )

        coords = torch.meshgrid(
            torch.linspace(0, 1, d),
            torch.linspace(0, 1, h),
            torch.linspace(0, 1, w),
            indexing="ij",
        )
        coords = torch.stack(coords, dim=-1).to(bbox.device)  # (d, h, w, 3)
        coords = coords.view(1, 1, d, h, w, 3)  # (1, 1, d, h, w, 3)
        coords = coords.repeat(B, L, 1, 1, 1, 1)  # (B, L, d, h, w, 3)

        # Scale and offset coords according to bbox
        top_left = bbox[:, :, 0, :].view(B, L, 1, 1, 1, 3)  # (B, L, 1, 1, 1, 3)
        bottom_right = bbox[:, :, 1, :].view(B, L, 1, 1, 1, 3)  # (B, L, 1, 1, 1, 3)
        coords = top_left + coords * (bottom_right - top_left)  # (B, L, d, h, w, 3)

        coords = rearrange(coords, "b l d h w c -> b (l d h w) c")
        return coords.to(bbox.device)

    def patch_embed(
        self, x: Tensor, bbox: Optional[tuple[Tensor, Tensor]] = None
    ) -> Tuple[Tensor, Tensor, Tensor]:
        """Process multi-level input through patch embedding.

        Input: x (B, L, C, D, H, W), bbox (B, L, 2, 3)
        Output: x (B, N, D), patches (B, N, C*D*H*W), coords (B, N, 3)
        """
        B, L, C, D, H, W = x.shape
        assert L == len(self.levels), "Number of levels must match"

        if bbox is not None:
            expected_shape = (B, L, 2, 3)
            if not (bbox.ndim == 4 and bbox.shape[:3] == (B, L, 2)):
                raise ValueError(
                    f"Bounding box must be of shape {expected_shape}: {bbox.shape}"
                )

        coords = self.get_patch_coords(B, L, D, H, W, bbox=bbox).to(x.device)

        # patchify assumes batch dimension is first, so we loop over levels
        x = x.transpose(0, 1)
        patches = tuple(self.patch_space_transform(self.patchify(_x)) for _x in x)
        x = tuple(pro(pat) for pro, pat in zip(self.proj, patches))

        if self.level_embed is not None:
            x = tuple(x + le for x, le in zip(x, self.level_embed))

        # back to B, L order
        x = torch.cat(x, dim=1)
        patches = torch.cat(patches, dim=1)

        return x, patches, coords

    def compute_features(self, x: Tensor, bbox: Optional[Tensor] = None) -> Tensor:
        """
        Compute features from a given input stack/bbox and return them in a spatially structured way.

        Args:
            x: Input tensor of shape (B, L, C, D, H, W)
            bbox: Optional bounding box coords

        Returns:
            feats: Features of shape (B, L, N, D', H', W') where H' = H // patch_size[0], W' = W // patch_size[1], D is the feature dimension
        """
        feats, _, _ = self(x, bbox=bbox)
        D, H, W = x.shape[-3], x.shape[-2], x.shape[-1]
        L = len(self.levels)
        d = D // self.patch_size[0]
        h = H // self.patch_size[1]
        w = W // self.patch_size[2]
        return rearrange(feats, "b (l d h w) n -> b l n d h w", l=L, h=h, w=w, d=d)


class MuViTEncoder4d(MuViTEncoder[Tuple[int, int, int]]):
    @classmethod
    @property
    def ndim(self) -> int:
        return 4

    def get_patch_coords(
        self,
        B: int,
        L: int,
        T: int,
        D: int,
        H: int,
        W: int,
        bbox: Optional[tuple[Tensor, Tensor]] = None,
    ) -> Tensor:
        """Get coordinates for each patch.

        Returns: (B,L*D*H*W,3) coordinates
        """
        t, d, h, w = (
            T // self.patch_size[0],
            D // self.patch_size[1],
            H // self.patch_size[2],
            W // self.patch_size[3],
        )

        if bbox is None:
            bbox = (
                [
                    [-lv * (T / 2 - 0.5), -lv * (D / 2 - 0.5), -lv * (H / 2 - 0.5), -lv * (W / 2 - 0.5)]
                    for lv in self.levels
                ],
                [
                    [lv * (T / 2 - 0.5), lv * (D / 2 - 0.5), lv * (H / 2 - 0.5), lv * (W / 2 - 0.5)]
                    for lv in self.levels
                ],
            )
            bbox = torch.tensor(bbox).permute(1, 0, 2)
            bbox = bbox.view(1, L, 2, self.ndim).repeat(B, 1, 1, 1)

        if not bbox.ndim == 4 and bbox.shape[:3] == (B, L, 4):
            raise ValueError(
                f"Bounding box must be of shape (B, L, 2, 4): {bbox.shape}"
            )
        coords = torch.meshgrid(
            torch.linspace(0, 1, t),
            torch.linspace(0, 1, d),
            torch.linspace(0, 1, h),
            torch.linspace(0, 1, w),
            indexing="ij",
        )
        coords = torch.stack(coords, dim=-1).to(bbox.device)  # (t, d, h, w, 4)
        coords = coords.view(1, 1, t, d, h, w, 4)  # (1, 1, t, d, h, w, 4)
        coords = coords.repeat(B, L, 1, 1, 1, 1, 1)  # (B, L, t, d, h, w, 4)

        # Scale and offset coords according to bbox
        top_left = bbox[:, :, 0, :].view(B, L, 1, 1, 1, 1, self.ndim)  # (B, L, 1, 1, 1, 1, 3)
        bottom_right = bbox[:, :, 1, :].view(B, L, 1, 1, 1, 1, self.ndim)  # (B, L, 1, 1, 1, 1, 3)
        coords = top_left + coords * (bottom_right - top_left)  # (B, L, t, d, h, w, 3)

        coords = rearrange(coords, "b l t d h w c -> b (l t d h w) c")
        return coords.to(bbox.device)

    def patch_embed(
        self, x: Tensor, bbox: Optional[tuple[Tensor, Tensor]] = None
    ) -> Tuple[Tensor, Tensor, Tensor]:
        """Process multi-level input through patch embedding.

        Input: x (B, L, C, T, D, H, W), bbox (B, L, 2, 4)
        Output: x (B, N, D), patches (B, N, C*T*D*H*W), coords (B, N, 4)
        """
        B, L, C, T, D, H, W = x.shape
        assert L == len(self.levels), "Number of levels must match"

        if bbox is not None:
            expected_shape = (B, L, 2, 4)
            if not (bbox.ndim == 4 and bbox.shape[:3] == (B, L, 2)):
                raise ValueError(
                    f"Bounding box must be of shape {expected_shape}: {bbox.shape}"
                )

        coords = self.get_patch_coords(B, L, T, D, H, W, bbox=bbox).to(x.device)

        # patchify assumes batch dimension is first, so we loop over levels
        x = x.transpose(0, 1)
        patches = tuple(self.patch_space_transform(self.patchify(_x)) for _x in x)
        x = tuple(pro(pat) for pro, pat in zip(self.proj, patches))

        if self.level_embed is not None:
            x = tuple(x + le for x, le in zip(x, self.level_embed))

        # back to B, L order
        x = torch.cat(x, dim=1)
        patches = torch.cat(patches, dim=1)

        return x, patches, coords

    def compute_features(self, x: Tensor, bbox: Optional[Tensor] = None) -> Tensor:
        """
        Compute features from a given input stack/bbox and return them in a spatially structured way.

        Args:
            x: Input tensor of shape (B, L, C, T, D, H, W)
            bbox: Optional bounding box coords

        Returns:
            feats: Features of shape (B, L, N, T', D', H', W') where H' = H // patch_size[0], W' = W // patch_size[1], D is the feature dimension
        """
        feats, _, _ = self(x, bbox=bbox)
        T, D, H, W = x.shape[-4], x.shape[-3], x.shape[-2], x.shape[-1]
        L = len(self.levels)
        t = T // self.patch_size[0]
        d = D // self.patch_size[1]
        h = H // self.patch_size[2]
        w = W // self.patch_size[3]
        return rearrange(feats, "b (l t d h w) n -> b l n t d h w", l=L, t=t, d=d, h=h, w=w)



import logging
import sys
from abc import ABC
from pathlib import Path
from typing import Literal, Optional

import torch
import torch.nn as nn
import yaml
from torch import Tensor
from x_transformers.x_transformers import Attention, FeedForward, RotaryEmbedding

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)

console_handler = logging.StreamHandler(sys.stdout)
console_handler.setLevel(logging.INFO)
formatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")
console_handler.setFormatter(formatter)
log.addHandler(console_handler)


class RotaryEmbeddingTrainable(RotaryEmbedding):
    def __init__(self, dim: int, base: int = 256):
        super().__init__(dim=dim, base=base)
        self.inv_freq = nn.Parameter(self.inv_freq)


def _split_rope_dims(dim: int, heads: int, ndim: int) -> list[int]:
    """Split head dimension for multi-dimensional RoPE embeddings."""
    dims = [dim // heads // ndim] * (ndim - 1)
    dims = [2 * (d // 2) for d in dims]  # Ensure even
    dims = [dim // heads - sum(dims)] + dims  # Residual in first dim
    assert min(dims) > 8, f"RoPE dimension too small: min={min(dims)}, need >8"
    return dims


def compute_rotary_emb(coords: Tensor, rotary_embs: nn.ModuleList) -> tuple[Tensor, Tensor]:
    """Compute rotary embeddings from coordinates."""
    freqs, scales = zip(*[r(coords[..., i]) for i, r in enumerate(rotary_embs)])
    freqs = torch.cat(freqs, dim=-1).contiguous()
    scales = scales[0]
    return freqs, scales


class TransformerLayer(nn.Module):
    def __init__(
        self,
        dim: int,
        rotary_dim: Optional[int],
        heads: int = 4,
        rotary_base: int = 10000,
        dropout: float = 0.0,
        rotary_pos_embs: Optional[nn.ModuleList] = None,
    ):
        """Initialize a transformer layer with attention and feed-forward components.

        Args:
            dim: Dimension of input/output tensors
            heads: Number of attention heads
            rotary_dim: Dimension for rotary embeddings, or None to disable
            rotary_base: Base for rotary embeddings
            dropout: Dropout probability for attention and feed-forward layers
            rotary_pos_embs: Optional pre-created rotary embeddings to share across layers
        """
        super().__init__()
        self.attn = Attention(
            dim=dim, heads=heads, dim_head=dim // heads, flash=True, dropout=dropout
        )
        self.norm1 = nn.LayerNorm(dim)
        self.norm2 = nn.LayerNorm(dim)
        self.ff = FeedForward(dim, dim, mult=2, dropout=dropout)

        if rotary_pos_embs is not None:
            self.rotary_pos_embs = rotary_pos_embs
        elif rotary_dim is not None:
            dims = [dim // heads // rotary_dim] * (rotary_dim - 1)
            dims = [2 * (d // 2) for d in dims]
            dims = [dim // heads - sum(dims)] + dims
            assert min(dims) > 8, "Rotary embedding dimension must be greater than 8"
            self.rotary_pos_embs = nn.ModuleList(
                [RotaryEmbeddingTrainable(dim=d, base=rotary_base) for d in dims]
            )
        else:
            self.rotary_pos_embs = None

    def forward(
        self,
        x: Tensor,
        level_idx: Tensor = None,
        coords: Tensor = None,
        attention_mode: Literal["all", "causal", "same", "random"] = "all",
        context: Tensor = None,
        context_level_idx: Tensor = None,
        context_coords: Tensor = None,
    ):
        """Apply transformer layer with optional rotary embeddings and level-based attention.

        If context is provided, performs cross-attention between x and context. Otherwise performs self-attention.
        Supports rotary embeddings and level-based attention masking.

        Args:
            x: Input tensor of shape (B, N, D)
            level_idx: Level indices for each token
            coords: Coordinates for rotary embeddings
            attention_mode: Type of attention masking ('all', 'causal', 'same', 'random')
            context: Optional context tensor for cross-attention. If provided, performs cross-attention between x and context
            context_level_idx: Level indices for context tokens
            context_coords: Coordinates for context rotary embeddings

        Returns:
            Transformed tensor of same shape as input
        """
        B, N, D = x.shape

        rotary_pos_emb = None
        if coords is not None and self.rotary_pos_embs is not None:
            assert coords.shape[:2] == x.shape[:2]
            rotary_pos_emb = compute_rotary_emb(coords, self.rotary_pos_embs)

        context_rotary_pos_emb = None
        if context_coords is not None and self.rotary_pos_embs is not None:
            assert context_coords.shape[:2] == context.shape[:2]
            context_rotary_pos_emb = compute_rotary_emb(context_coords, self.rotary_pos_embs)

        if context is not None:
            assert context.shape[0] == B
            assert context.shape[2] == D
            NC = context.shape[1]
        else:
            NC = N

        if context_level_idx is None:
            context_level_idx = level_idx

        if level_idx is not None:
            if attention_mode == "all":
                mask = None
            elif attention_mode == "causal":
                mask = level_idx.view(B, 1, N, 1) >= context_level_idx.view(B, 1, 1, NC)
            elif attention_mode == "same":
                mask = level_idx.view(B, 1, NC, 1) == context_level_idx.view(
                    B, 1, 1, NC
                )
            elif attention_mode == "random":
                # Create random boolean mask for each batch and head
                mask = torch.argsort(torch.rand(N, NC, device=x.device), dim=1) == 0
            else:
                raise ValueError(f"Invalid attention mode: {attention_mode}")
        else:
            mask = None

        x = x + self.attn(
            self.norm1(x),
            attn_mask=mask,
            rotary_pos_emb=rotary_pos_emb,
            context=context,
            context_rotary_pos_emb=context_rotary_pos_emb,
        )
        x = x + self.ff(self.norm2(x))
        return x


class TransformerDecoderLayer(nn.Module):
    def __init__(
        self,
        dim: int,
        heads: int = 4,
        rotary_dim: Optional[int] = None,
        rotary_dim_context: Optional[int] = None,
        rotary_base: int = 10000,
        dropout: float = 0.0,
        ff_mult: int = 4,
        with_cross_attn: bool = True,
        rotary_pos_embs: Optional[nn.ModuleList] = None,
        rotary_pos_embs_context: Optional[nn.ModuleList] = None,
    ):
        """Transformer decoder layer with self-attention, cross-attention, and feed-forward.

        Args:
            dim: Feature dimension
            heads: Number of attention heads
            rotary_dim: RoPE dimensionality for queries (None to disable)
            rotary_dim_context: RoPE dimensionality for context keys (None to disable)
            rotary_base: Base for RoPE
            dropout: Dropout rate
            ff_mult: Feedforward expansion factor (default 4 to match classical transformers)
            with_cross_attn: Whether to include cross-attention layer
            rotary_pos_embs: Pre-created RoPE modules for queries (shared across layers)
            rotary_pos_embs_context: Pre-created RoPE modules for context (shared across layers)
        """
        super().__init__()

        self.with_cross_attn = with_cross_attn

        # Self-attention
        self.self_attn = Attention(
            dim=dim, heads=heads, dim_head=dim // heads, flash=True, dropout=dropout
        )
        self.norm1 = nn.LayerNorm(dim)

        # Cross-attention (optional)
        if with_cross_attn:
            self.cross_attn = Attention(
                dim=dim, heads=heads, dim_head=dim // heads, flash=True, dropout=dropout
            )
            self.norm2 = nn.LayerNorm(dim)
            self.norm3 = nn.LayerNorm(dim)
        else:
            self.norm2 = nn.LayerNorm(dim)

        # Feed-forward
        self.ff = FeedForward(dim, dim, mult=ff_mult, dropout=dropout)

        # Rotary embeddings for QUERIES
        if rotary_pos_embs is not None:
            self.rotary_pos_embs = rotary_pos_embs
        elif rotary_dim is not None:
            dims = _split_rope_dims(dim, heads, rotary_dim)
            self.rotary_pos_embs = nn.ModuleList(
                [RotaryEmbeddingTrainable(dim=d, base=rotary_base) for d in dims]
            )
        else:
            self.rotary_pos_embs = None

        # Rotary embeddings for CONTEXT KEYS
        if rotary_pos_embs_context is not None:
            self.rotary_pos_embs_context = rotary_pos_embs_context
        elif rotary_dim_context is not None:
            dims_context = _split_rope_dims(dim, heads, rotary_dim_context)
            self.rotary_pos_embs_context = nn.ModuleList(
                [RotaryEmbeddingTrainable(dim=d, base=rotary_base) for d in dims_context]
            )
        else:
            self.rotary_pos_embs_context = None

    def forward(
        self,
        x: Tensor,
        context: Optional[Tensor] = None,
        coords: Optional[Tensor] = None,
        context_coords: Optional[Tensor] = None,
        level_idx: Optional[Tensor] = None,
        context_level_idx: Optional[Tensor] = None,
        attention_mode: Literal["all", "causal", "same", "random"] = "all",
    ):
        """Apply decoder layer: self-attention -> cross-attention -> feed-forward.

        Args:
            x: Queries (B, N, D)
            context: Context for cross-attention (B, M, D), None to skip cross-attention
            coords: Coordinates for query RoPE (B, N, ndim)
            context_coords: Coordinates for context key RoPE (B, M, ndim)
            level_idx: Level indices for queries
            context_level_idx: Level indices for context
            attention_mode: Attention masking mode

        Returns:
            Transformed tensor of same shape as input
        """
        B, N, D = x.shape

        # === SELF-ATTENTION ===

        # Compute RoPE for queries
        query_rotary_emb = None
        if coords is not None and self.rotary_pos_embs is not None:
            assert coords.shape[:2] == x.shape[:2]
            query_rotary_emb = compute_rotary_emb(coords, self.rotary_pos_embs)

        # Self-attention mask
        self_attn_mask = None
        if level_idx is not None:
            if attention_mode == "all":
                self_attn_mask = None
            elif attention_mode == "causal":
                self_attn_mask = level_idx.view(B, 1, N, 1) >= level_idx.view(B, 1, 1, N)
            elif attention_mode == "same":
                self_attn_mask = level_idx.view(B, 1, N, 1) == level_idx.view(B, 1, 1, N)
            elif attention_mode == "random":
                self_attn_mask = torch.argsort(torch.rand(N, N, device=x.device), dim=1) == 0
            else:
                raise ValueError(f"Invalid attention mode: {attention_mode}")

        # Apply self-attention
        x = x + self.self_attn(
            self.norm1(x),
            attn_mask=self_attn_mask,
            rotary_pos_emb=query_rotary_emb,
        )

        # === CROSS-ATTENTION (if enabled and context provided) ===

        if self.with_cross_attn and context is not None:
            M = context.shape[1]
            assert context.shape[0] == B and context.shape[2] == D

            # Compute RoPE for context keys
            context_rotary_emb = None
            if context_coords is not None and self.rotary_pos_embs_context is not None:
                assert context_coords.shape[:2] == context.shape[:2]
                context_rotary_emb = compute_rotary_emb(
                    context_coords, self.rotary_pos_embs_context
                )

            # Cross-attention mask
            cross_attn_mask = None
            if level_idx is not None and context_level_idx is not None:
                if attention_mode == "all":
                    cross_attn_mask = None
                elif attention_mode == "causal":
                    cross_attn_mask = level_idx.view(B, 1, N, 1) >= context_level_idx.view(B, 1, 1, M)
                elif attention_mode == "same":
                    cross_attn_mask = level_idx.view(B, 1, N, 1) == context_level_idx.view(B, 1, 1, M)

            # Apply cross-attention
            x = x + self.cross_attn(
                self.norm2(x),
                context=context,
                attn_mask=cross_attn_mask,
                rotary_pos_emb=None,  # Queries don't get RoPE in cross-attention
                context_rotary_pos_emb=context_rotary_emb,
            )

            x = x + self.ff(self.norm3(x))
        else:
            # No cross-attention, just feed-forward
            x = x + self.ff(self.norm2(x))

        return x


class SaveableModel(nn.Module, ABC):
    """Base class for models that can be saved and loaded from disk."""

    def __init__(self):
        super().__init__()
        self._config = {}

    def save(self, path: str | Path, overwrite: bool = False):
        """Save model configuration and state to a folder.

        Args:
            path: Path to save the model
            overwrite: Whether to overwrite existing files
        """
        log.info(f"Saving model to {path}")
        path = Path(path)
        if path.exists() and not overwrite:
            raise ValueError(
                f"Path already exists: {path} (set overwrite=True to overwrite)"
            )
        path.mkdir(parents=True, exist_ok=True)
        with open(path / "config.yaml", "w") as f:
            yaml.dump(self._config, f)
        torch.save(self.state_dict(), path / "model.pth")

    @classmethod
    def from_folder(cls, path: str | Path, device: str | torch.device = "cpu"):
        """Load model from a saved folder.

        Args:
            path: Path to the saved model folder
            device: Device to load the model on

        Returns:
            Loaded model instance
        """
        path = Path(path)
        with open(path / "config.yaml", "r") as f:
            config = yaml.load(f, Loader=yaml.FullLoader)
        if hasattr(cls, "ndim") and "ndim" in config.keys():
            if config["ndim"] != cls.ndim:
                raise ValueError(f"Expected ndim={cls.ndim}, but got {config['ndim']}. Are you sure you're instantiating the correct class?")
        if "ndim" in config:
            config.pop("ndim", None)

        # Backward compatibility: convert old use_rotary_embed to rotary_mode
        if "use_rotary_embed" in config:
            if "rotary_mode" in config:
                raise ValueError(
                    "Config contains both 'use_rotary_embed' (deprecated) and 'rotary_mode'. "
                    "Please use only 'rotary_mode'."
                )
            config["rotary_mode"] = "per_layer" if config["use_rotary_embed"] else "none"
            config.pop("use_rotary_embed")

        model = cls(**config).to(device)
        model.load_state_dict(torch.load(path / "model.pth", map_location=device))
        return model
